import math
from abc import abstractmethod

from pymoo.core.termination import Termination


class DeltaToleranceTermination(Termination):

    def __init__(self, tol, n_skip=0):
        super().__init__()

        # the tolerance threshold the difference (delta) to be under
        assert tol >= 0
        self.tol = tol

        # the previous values to calculate the difference
        self.data = None

        # a counter of update calls
        self.counter = 0

        # whether some updates should be skipped
        self.n_skip = n_skip

    def _update(self, algorithm):

        # the object from the previous iteration
        prev = self.data

        # and the one from the current iteration
        current = self._data(algorithm)

        # if there is no previous element to use
        if prev is None:
            perc = 0.0

        elif self.counter > 0 and self.counter % (self.n_skip + 1) != 0:
            perc = self.perc

        else:
            tol = self.tol
            delta = self._delta(prev, current)

            if delta <= tol:
                return 1.0
            else:
                v = (delta - tol)
                perc = 1 / (1 + v)

        # remember the data from the current iteration and set it to data
        self.data = current

        # increase the function update counter
        self.counter += 1

        return perc

    @abstractmethod
    def _delta(self, prev, current):
        pass

    @abstractmethod
    def _data(self, algorithm):
        pass



import numpy as np

from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot
from pymoo.util.misc import set_if_none_from_tuples
from pymoo.visualization.util import plot_axes_lines, plot_axis_labels, plot_polygon, get_circle_points, \
    plot_radar_line, equal_axis, no_ticks, parse_bounds, normalize


class Radar(Plot):

    def __init__(self,
                 normalize_each_objective=True,
                 n_partitions=3,
                 point_style={},
                 **kwargs):

        """
        Radar Plot

        Parameters
        ----------------
        normalize_each_objective : bool
            Whether each objective is normalized. Otherwise the inner and outer bound is plotted.
        point_style : dict
            The style being used to visualize the points
        n_partitions : int
            Number of partitions to show in the radar.
        reverse : {reverse}
        axis_style : {axis_style}
        labels : {labels}

        Other Parameters
        ----------------
        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}
        """

        super().__init__(**kwargs)
        self.normalize_each_objective = normalize_each_objective
        self.n_partitions = n_partitions

        self.point_style = point_style
        set_if_none_from_tuples(self.point_style, ("s", 15))

        set_if_none_from_tuples(self.axis_style, ("color", "black"), ("linewidth", 0.5), ("alpha", 0.75))

    def _plot(self, ax, _F, inner, outer, kwargs):

        set_if_none_from_tuples(kwargs, ("alpha", 0.5))

        # equal axis length and no ticks
        equal_axis(ax)
        no_ticks(ax)

        # draw the axis lines and labels
        plot_axes_lines(ax, outer, extend_factor=1.0, **self.axis_style)
        plot_axis_labels(ax, outer, self.get_labels(), margin=0.015, **self.axis_label_style)

        # plot the outer radar line and the inner polygon
        plot_radar_line(ax, outer, **self.axis_style)
        plot_polygon(ax, inner)

        # find the corresponding point
        _F = inner + _F[:, None] * (outer - inner)

        # plot the points and no polygon
        ax.scatter(_F[:, 0], _F[:, 1], **self.point_style)
        plot_polygon(ax, _F, **kwargs)

    def _do(self):

        if self.bounds is None:
            raise Exception("The boundaries must be provided.")

        _F = np.row_stack([e[0] for e in self.to_plot])
        if np.any(_F < self.bounds[0]) or np.any(_F > self.bounds[1]):
            raise Exception(
                "Points out of the boundaries exist! Please make sure the boundaries are indeed boundaries.")

        n_rows = len(self.to_plot)
        n_cols = max([len(e[0]) for e in self.to_plot])
        self.init_figure(n_rows=n_rows, n_cols=n_cols, force_axes_as_matrix=True)

        # normalize the input
        bounds = parse_bounds(self.bounds, self.n_dim)
        to_plot_norm = normalize(self.to_plot, bounds, reverse=self.reverse)

        # get the endpoints of circle
        V = get_circle_points(self.n_dim)

        if self.normalize_each_objective:
            inner = np.zeros((self.n_dim, 1)) * V
            outer = np.ones((self.n_dim, 1)) * V
        else:
            inner = bounds[[0]].T * V
            outer = (bounds[[1]].T * V) / bounds[1].max()

        for k, (F, kwargs) in enumerate(to_plot_norm):

            for j, _F in enumerate(F):
                self._plot(self.ax[k, j], _F, inner, outer, kwargs)


parse_doc_string(Radar.__init__)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation


def plot(*args, show=True, labels=None, no_fill=False, **kwargs):
    F = args[0]

    if F.ndim == 1:
        print("Cannot plot a one dimensional array.")
        return

    n_dim = F.shape[1]

    if n_dim == 2:
        ret = plot_2d(*args, labels=labels, no_fill=no_fill, **kwargs)
    elif n_dim == 3:
        ret = plot_3d(*args, labels=labels, no_fill=no_fill, **kwargs)
    else:
        print("Cannot plot a %s dimensional array." % n_dim)
        return

    if labels:
        plt.legend()

    if show:
        plt.show()

    return ret


def plot_3d(*args, no_fill=False, labels=None, **kwargs):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    for i, F in enumerate(args):

        if no_fill:
            kwargs["s"] = 20
            kwargs["marker"] = '.'
            kwargs["facecolors"] = (0, 0, 0, 0)
            kwargs["edgecolors"] = 'r'

        if labels:
            ax.scatter(F[:, 0], F[:, 1], F[:, 2], label=labels[i], **kwargs)
        else:
            ax.scatter(F[:, 0], F[:, 1], F[:, 2], **kwargs)

    return ax


def plot_2d(*args, labels=None, no_fill=False):
    if no_fill:
        kwargs = dict(
            s=20,
            facecolors='none',
            edgecolors='r'
        )
    else:
        kwargs = {}

    for i, F in enumerate(args):
        if labels:
            plt.scatter(F[:, 0], F[:, 1], label=labels[i], **kwargs)
        else:
            plt.scatter(F[:, 0], F[:, 1], **kwargs)


def animate(path_to_file, H, problem=None, func_iter=None, plot_min=None, plot_max=None):
    if H.ndim != 3 or H.shape[2] != 2:
        print("Can only animate a two dimensional set of arrays.")
        return

    fig = plt.figure()
    ax = plt.gca()

    # plot the pareto front if it is known for the problem
    if problem is not None:
        pf = problem.pareto_front()
        plt.scatter(pf[:, 0], pf[:, 1], label='Pareto Front', s=60, facecolors='none', edgecolors='r')

    # plot the initial population
    _F = H[0, :, :]
    scat = plt.scatter(_F[:, 0], _F[:, 1])
    plt.title("0")

    if func_iter is not None:
        func_iter(ax, H[0])

    # the update method
    def update(n):
        _F = H[n, :, :]
        scat.set_offsets(_F)

        # get the bounds for plotting and add padding
        min = np.min(_F, axis=0) - 0.1
        max = np.max(_F, axis=0) + 0.1

        # set the scatter object with padding
        ax.set_xlim(min[0], max[0])
        ax.set_ylim(min[1], max[1])

        if func_iter is not None:
            func_iter(ax, H[n])

        plt.title(n)

    # create the animation
    ani = animation.FuncAnimation(fig, update, frames=H.shape[0])

    # write the file
    Writer = animation.writers['ffmpeg']
    writer = Writer(fps=6, bitrate=1800)
    ani.save(path_to_file, writer=writer)

    print("Saving: ", path_to_file)


def plot_problem_surface(problem, n_samples, plot_type="wireframe", cmap="summer", show=True, return_figure=False):
    try:
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
    except:
        raise Exception("Please install 'matplotlib' to use the plotting functionality.")

    fig = plt.figure()

    if problem.n_var == 1 and problem.n_obj == 1:

        X = np.linspace(problem.xl[0], problem.xu[0], num=n_samples)[:, None]
        Y = problem.evaluate(X, return_values_of=["F"])
        ax = plt.plot(X, Y)

    elif problem.n_var == 2 and problem.n_obj == 1:

        X_range = np.linspace(problem.xl[0], problem.xu[0], num=n_samples)
        Y_range = np.linspace(problem.xl[1], problem.xu[1], num=n_samples)
        X, Y = np.meshgrid(X_range, Y_range)

        A = np.zeros((n_samples * n_samples, 2))
        counter = 0
        for i, x in enumerate(X_range):
            for j, y in enumerate(Y_range):
                A[counter, 0] = x
                A[counter, 1] = y
                counter += 1

        F = np.reshape(problem.evaluate(A, return_values_of=["F"]), (n_samples, n_samples))

        # Plot the surface.
        if plot_type == "wireframe":
            ax = fig.add_subplot(111, projection='3d')
            ax.plot_wireframe(X, Y, F)
        elif plot_type == "contour":
            CS = plt.contour(X, Y, F)
            plt.clabel(CS, inline=1, fontsize=10)
        elif plot_type == "wireframe+contour":
            ax = fig.add_subplot(111, projection="3d")
            ax.plot_surface(X, Y, F, cmap=cmap, rstride=1, cstride=1)
            ax.contour(X, Y, F, 10, linestyles="solid", offset=-1)
            ax.set_xlabel("$x_1$")
            ax.set_ylabel("$x_2$")
            ax.set_zlabel("$f(x)$")
            ax.view_init(45, 45)
        else:
            raise Exception("Unknown plotting method.")


    else:
        raise Exception("Can only plot single with less than two variables and one objective.")

    if show:
        plt.tight_layout()
        plt.show()

    if return_figure:
        return fig, ax



from abc import abstractmethod

import numpy as np

from pymoo.core.callback import Callback
from pymoo.indicators.igd import IGD
from pymoo.indicators.igd_plus import IGDPlus


class Recorder(Callback):

    def __init__(self, nth_evals=None) -> None:
        super().__init__()
        self.data = []
        self.nth_evals = nth_evals
        self.rec_n_evals = 0

    def notify(self, algorithm, **kwargs):

        if self.nth_evals is None:
            self.data.append(self.save(algorithm))

        else:
            n_evals = algorithm.evaluator.n_eval

            if n_evals >= self.rec_n_evals:
                self.data.append(self.save(algorithm))
                self.rec_n_evals = (1 + (n_evals // self.nth_evals)) * self.nth_evals

    @abstractmethod
    def save(self, algorithm):
        pass

    def get(self, *args, as_array=True):
        ret = []
        for arg in args:
            e = [entry.get(arg) for entry in self.data]
            if as_array:
                e = np.array(e)
            ret.append(e)
        return tuple(ret)


class DefaultSingleObjectiveRecorder(Recorder):

    def save(self, algorithm):
        n_evals = algorithm.evaluator.n_eval

        opt = algorithm.opt
        _feas, _cv, _f = opt.get("feas", "cv", "f")

        cv = _cv.min()

        if np.any(_feas):
            f = _f[_feas].min()
        else:
            f = np.inf

        fgap = np.inf
        try:
            pf = algorithm.problem.pareto_front()
        except:
            pf = None

        if pf is not None:
            fgap = f - pf.min()

        return dict(n_evals=n_evals, cv=cv, f=f, fgap=fgap)


class DefaultMultiObjectiveRecorder(Recorder):

    def save(self, algorithm):

        igd, igd_plus = np.inf, np.inf

        opt = algorithm.opt

        # get all optimal solutions that are feasible
        feas_opt = opt[opt.get("feas")]

        if len(feas_opt) > 0:

            try:
                pf = algorithm.problem.pareto_front()
            except:
                pf = None

            if pf is not None:
                F = feas_opt.get("F")
                igd = IGD(pf=pf, zero_to_one=True).do(F)
                igd_plus = IGDPlus(pf=pf, zero_to_one=True).do(F)

        return {
            "n_evals": algorithm.evaluator.n_eval,
            "cv": opt.get("cv").min(),
            "igd": igd,
            "igd+": igd_plus,
        }



import sys

import numpy as np
from scipy import special

from pymoo.util.misc import find_duplicates, cdist


# =========================================================================================================
# Model
# =========================================================================================================


def default_ref_dirs(m):
    if m == 1:
        return np.array([[1.0]])
    elif m == 2:
        return UniformReferenceDirectionFactory(m, n_partitions=99).do()
    elif m == 3:
        return UniformReferenceDirectionFactory(m, n_partitions=12).do()
    else:
        raise Exception("No default reference directions for more than 3 objectives. Please provide them directly:"
                        "https://pymoo.org/misc/reference_directions.html")


class ReferenceDirectionFactory:

    def __init__(self, n_dim, scaling=None, lexsort=True, verbose=False, seed=None, **kwargs) -> None:
        super().__init__()
        self.n_dim = n_dim
        self.scaling = scaling
        self.lexsort = lexsort
        self.verbose = verbose
        self.seed = seed

    def __call__(self):
        return self.do()

    def do(self):

        # set the random seed if it is provided
        if self.seed is not None:
            np.random.seed(self.seed)

        if self.n_dim == 1:
            return np.array([[1.0]])
        else:

            val = self._do()
            if isinstance(val, tuple):
                ref_dirs, other = val[0], val[1:]
            else:
                ref_dirs = val

            if self.scaling is not None:
                ref_dirs = scale_reference_directions(ref_dirs, self.scaling)

            # do ref_dirs is desired
            if self.lexsort:
                I = np.lexsort([ref_dirs[:, j] for j in range(ref_dirs.shape[1])][::-1])
                ref_dirs = ref_dirs[I]

            return ref_dirs

    def _do(self):
        return None


# =========================================================================================================
# Das Dennis Reference Directions (Uniform)
# =========================================================================================================


def get_number_of_uniform_points(n_partitions, n_dim):
    """
    Returns the number of uniform points that can be created uniformly.
    """
    return int(special.binom(n_dim + n_partitions - 1, n_partitions))


def get_partition_closest_to_points(n_points, n_dim):
    """
    Returns the corresponding partition number which create the desired number of points
    or less!
    """

    if n_dim == 1:
        return 0

    n_partitions = 1
    _n_points = get_number_of_uniform_points(n_partitions, n_dim)
    while _n_points <= n_points:
        n_partitions += 1
        _n_points = get_number_of_uniform_points(n_partitions, n_dim)
    return n_partitions - 1


def das_dennis(n_partitions, n_dim):
    if n_partitions == 0:
        return np.full((1, n_dim), 1 / n_dim)
    else:
        ref_dirs = []
        ref_dir = np.full(n_dim, np.nan)
        das_dennis_recursion(ref_dirs, ref_dir, n_partitions, n_partitions, 0)
        return np.concatenate(ref_dirs, axis=0)


def das_dennis_recursion(ref_dirs, ref_dir, n_partitions, beta, depth):
    if depth == len(ref_dir) - 1:
        ref_dir[depth] = beta / (1.0 * n_partitions)
        ref_dirs.append(ref_dir[None, :])
    else:
        for i in range(beta + 1):
            ref_dir[depth] = 1.0 * i / (1.0 * n_partitions)
            das_dennis_recursion(ref_dirs, np.copy(ref_dir), n_partitions, beta - i, depth + 1)


class UniformReferenceDirectionFactory(ReferenceDirectionFactory):

    def __init__(self, n_dim, scaling=None, n_points=None, n_partitions=None, **kwargs) -> None:
        super().__init__(n_dim, scaling=scaling, **kwargs)

        if n_points is not None:
            n_partitions = get_partition_closest_to_points(n_points, n_dim)
            results_in = get_number_of_uniform_points(n_partitions, n_dim)

            # the number of points are not matching to any partition number
            if results_in != n_points:
                results_in_next = get_number_of_uniform_points(n_partitions + 1, n_dim)
                raise Exception("The number of points (n_points = %s) can not be created uniformly.\n"
                                "Either choose n_points = %s (n_partitions = %s) or "
                                "n_points = %s (n_partitions = %s)." %
                                (n_points, results_in, n_partitions, results_in_next, n_partitions + 1))

            self.n_partitions = n_partitions

        elif n_partitions is not None:
            self.n_partitions = n_partitions

        else:
            raise Exception("Either provide number of partitions or number of points.")

    def _do(self):
        return das_dennis(self.n_partitions, self.n_dim)


# =========================================================================================================
# Multi Layer
# =========================================================================================================


class MultiLayerReferenceDirectionFactory:

    def __init__(self, *args) -> None:
        self.layers = []
        self.layers.extend(args)

    def __call__(self):
        return self.do()

    def add_layer(self, *args):
        self.layers.extend(args)

    def do(self):
        ref_dirs = []
        for factory in self.layers:
            ref_dirs.append(factory)
        ref_dirs = np.concatenate(ref_dirs, axis=0)
        is_duplicate = find_duplicates(ref_dirs)
        return ref_dirs[np.logical_not(is_duplicate)]


# =========================================================================================================
# Util
# =========================================================================================================


def sample_on_unit_simplex(n_points, n_dim, unit_simplex_mapping="kraemer"):
    if unit_simplex_mapping == "sum":
        rnd = map_onto_unit_simplex(np.random.random((n_points, n_dim)), "sum")

    elif unit_simplex_mapping == "kraemer":
        rnd = map_onto_unit_simplex(np.random.random((n_points, n_dim)), "kraemer")

    elif unit_simplex_mapping == "das-dennis":
        n_partitions = get_partition_closest_to_points(n_points, n_dim)
        rnd = UniformReferenceDirectionFactory(n_dim, n_partitions=n_partitions).do()

    else:
        raise Exception("Please define a valid sampling on unit simplex strategy!")

    return rnd


def map_onto_unit_simplex(rnd, method):
    n_points, n_dim = rnd.shape

    if method == "sum":
        ret = rnd / rnd.sum(axis=1)[:, None]

    elif method == "kraemer":
        M = sys.maxsize

        rnd *= M
        rnd = rnd[:, :n_dim - 1]
        rnd = np.column_stack([np.zeros(n_points), rnd, np.full(n_points, M)])

        rnd = np.sort(rnd, axis=1)

        ret = np.full((n_points, n_dim), np.nan)
        for i in range(1, n_dim + 1):
            ret[:, i - 1] = rnd[:, i] - rnd[:, i - 1]
        ret /= M

    else:
        raise Exception("Invalid unit simplex mapping!")

    return ret


def scale_reference_directions(ref_dirs, scaling):
    return ref_dirs * scaling + ((1 - scaling) / ref_dirs.shape[1])


def select_points_with_maximum_distance(X, n_select, selected=[]):
    n_points, n_dim = X.shape

    # calculate the distance matrix
    D = cdist(X, X)

    # if no selection provided pick randomly in the beginning
    if len(selected) == 0:
        selected = [np.random.randint(len(X))]

    # create variables to store what selected and what not
    not_selected = [i for i in range(n_points) if i not in selected]

    # remove unnecessary points
    dist_to_closest_selected = D[:, selected].min(axis=1)

    # now select the points until sufficient ones are found
    while len(selected) < n_select:
        # find point that has the maximum distance to all others
        index_in_not_selected = dist_to_closest_selected[not_selected].argmax()
        I = not_selected[index_in_not_selected]

        # add the closest distance to selected point
        is_closer = D[I] < dist_to_closest_selected
        dist_to_closest_selected[is_closer] = D[I][is_closer]

        # add it to the selected and remove from not selected
        selected.append(I)
        not_selected = np.delete(not_selected, index_in_not_selected)

    return selected



import logging

from optuna.logging import get_logger

from pymoo.util.optimum import filter_optimum

try:
    import optuna
    from optuna.samplers import TPESampler
except:
    raise Exception("Please install optuna: pip install optuna")

from pymoo.core.algorithm import Algorithm
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.variable import Real, Integer, Choice, Binary
from pymoo.util.display.single import SingleObjectiveOutput


class Optuna(Algorithm):

    def __init__(self, sampler=None, output=SingleObjectiveOutput(), **kwargs):
        super().__init__(output=output, **kwargs)
        self.sampler = sampler

    def _setup(self, problem, **kwargs):

        sampler = self.sampler
        if sampler is None:
            sampler = TPESampler(seed=self.seed)

        # that disables the warning in the very beginning
        get_logger('optuna.storages._in_memory').disabled = True

        # create a new study
        self.study = optuna.create_study(study_name=f"Study@{id(self)}", sampler=sampler, direction='minimize')

        # the current trial for an individual
        self.trial = None

    def _infill(self):
        self.trial = self.study.ask()

        vars = self.problem.vars
        assert vars is not None, "Optuna needs explicitly defined variables."

        x = {}
        for name, param in vars.items():
            if isinstance(param, Real):
                lower, upper = param.bounds
                v = self.trial.suggest_float(name, lower, upper)
            elif isinstance(param, Integer):
                lower, upper = param.bounds
                v = self.trial.suggest_int(name, lower, upper)
            elif isinstance(param, Choice):
                options = param.options
                v = self.trial.suggest_categorical(name, options)
            elif isinstance(param, Binary):
                v = self.trial.suggest_categorical(name, [False, True])
            else:
                raise Exception("Type not supported yet.")
            x[name] = v

        return Individual(X=x)

    def _advance(self, infills=None, **kwargs):
        self.pop = Population.create(infills)
        self.study.tell(self.trial, infills.f)

    def _initialize_infill(self):
        return self._infill()

    def _initialize_advance(self, **kwargs):
        return self._advance(**kwargs)

    def _set_optimum(self):
        pop = self.pop
        if self.opt is not None:
            pop = Population.merge(self.opt, pop)
        self.opt = filter_optimum(pop, least_infeasible=True)






from copy import deepcopy


class Meta(object):

    def __init__(self, object, copy=True, clazz=None):
        if clazz is None:
            clazz = self.__class__

        wrapped = object
        if copy:
            wrapped = deepcopy(wrapped)

        self.__class__ = type(clazz.__name__,
                              tuple([clazz] + wrapped.__class__.mro()),
                              {})

        self.__dict__ = wrapped.__dict__
        self.__object__ = object
        self.__super__ = wrapped




import numpy as np

from pymoo.core.crossover import Crossover


def random_sequence(n):
    start, end = np.sort(np.random.choice(n, 2, replace=False))
    return tuple([start, end])


def ox(receiver, donor, seq=None, shift=False):
    """
    The Ordered Crossover (OX) as explained in http://www.dmi.unict.it/mpavone/nc-cs/materiale/moscato89.pdf.

    Parameters
    ----------
    receiver : numpy.array
        The receiver of the sequence. The array needs to be repaired after the donation took place.
    donor : numpy.array
        The donor of the sequence.
    seq : tuple (optional)
        Tuple with two problems defining the start and the end of the sequence. Please note in our implementation
        the end of the sequence is included. The sequence is randomly chosen if not provided.

    shift : bool
        Whether during the repair the receiver should be shifted or not. Both version of it can be found in the
        literature.

    Returns
    -------

    y : numpy.array
        The offspring which was created by the ordered crossover.

    """
    assert len(donor) == len(receiver)

    # the sequence which shall be use for the crossover
    seq = seq if not None else random_sequence(len(receiver))
    start, end = seq

    # the donation and a set of it to allow a quick lookup
    donation = np.copy(donor[start:end + 1])
    donation_as_set = set(donation)

    # the final value to be returned
    y = []

    for k in range(len(receiver)):

        # do the shift starting from the swapped sequence - as proposed in the paper
        i = k if not shift else (start + k) % len(receiver)
        v = receiver[i]

        if v not in donation_as_set:
            y.append(v)

    # now insert the donation at the right place
    y = np.concatenate([y[:start], donation, y[start:]]).astype(copy=False, dtype=int)

    return y


class OrderCrossover(Crossover):

    def __init__(self, shift=False, **kwargs):
        super().__init__(2, 2, **kwargs)
        self.shift = shift

    def _do(self, problem, X, **kwargs):
        _, n_matings, n_var = X.shape
        Y = np.full((self.n_offsprings, n_matings, n_var), -1, dtype=int)

        for i in range(n_matings):
            a, b = X[:, i, :]
            n = len(a)

            # define the sequence to be used for crossover
            start, end = random_sequence(n)

            Y[0, i, :] = ox(a, b, seq=(start, end), shift=self.shift)
            Y[1, i, :] = ox(b, a, seq=(start, end), shift=self.shift)

        return Y



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class OmniTest(Problem):
    """
    The Omni-test problem proposed by Deb in [1].

    Parameters
    ----------
    n_var: number of decision variables

    References
    ----------
    [1] Deb, K., Tiwari, S. "Omni-optimizer: A generic evolutionary algorithm for single and multi-objective optimization"
    """
    def __init__(self, n_var=2):
        assert (n_var >= 2), "The dimension of the decision space should at least be 2!"
        super().__init__(
            n_var=n_var, n_obj=2, vtype=float, xl=np.full(n_var, 0), xu=np.full(n_var, 6)
        )

    def _evaluate(self, X, out, *args, **kwargs):
        F1 = anp.sum(anp.sin(anp.pi * X), axis=1)
        F2 = anp.sum(anp.cos(anp.pi * X), axis=1)
        out["F"] = anp.vstack((F1, F2)).T

    def _calc_pareto_set(self, n_pareto_points=500):
        # The Omni-test problem has 3^D Pareto subsets
        num_ps = int(3 ** self.n_var)
        h = int(n_pareto_points / num_ps)
        PS = np.zeros((num_ps * h, self.n_var))

        candidates = np.array([np.linspace(2 * m + 1, 2 * m + 3 / 2, h) for m in range(3)])
        # generate combination indices
        candidates_indices = [[0, 1, 2] for _ in range(self.n_var)]
        a = np.meshgrid(*candidates_indices)
        combination_indices = np.array(a).T.reshape(-1, self.n_var)
        # generate 3^D combinations
        for i in range(num_ps):
            PS[i * h:i * h + h, :] = candidates[combination_indices[i]].T
        return PS

    def _calc_pareto_front(self, n_pareto_points=500):
        PS = self._calc_pareto_set(n_pareto_points)
        return self.evaluate(PS, return_values_of=["F"])



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.variable import Real, get
from pymoo.operators.repair.bounds_repair import repair_clamp


# ---------------------------------------------------------------------------------------------------------
# Function
# ---------------------------------------------------------------------------------------------------------


def cross_sbx(X, xl, xu, eta, prob_var, prob_bin, eps=1.0e-14):
    n_parents, n_matings, n_var = X.shape

    # the probability of a crossover for each of the variables
    cross = np.random.random((n_matings, n_var)) < prob_var

    # when solutions are too close -> do not apply sbx crossover
    too_close = np.abs(X[0] - X[1]) <= eps

    # disable if two individuals are already too close
    cross[too_close] = False

    # disable crossover when lower and upper bound are identical
    cross[:, xl == xu] = False

    # assign y1 the smaller and y2 the larger value
    y1 = np.min(X, axis=0)[cross]
    y2 = np.max(X, axis=0)[cross]

    # mask all the values that should be crossovered
    _xl = np.repeat(xl[None, :], n_matings, axis=0)[cross]
    _xu = np.repeat(xu[None, :], n_matings, axis=0)[cross]
    eta = eta.repeat(n_var, axis=1)[cross]
    prob_bin = prob_bin.repeat(n_var, axis=1)[cross]

    # random values for each individual
    rand = np.random.random(len(eta))

    def calc_betaq(beta):
        alpha = 2.0 - np.power(beta, -(eta + 1.0))

        mask, mask_not = (rand <= (1.0 / alpha)), (rand > (1.0 / alpha))

        betaq = np.zeros(mask.shape)
        betaq[mask] = np.power((rand * alpha), (1.0 / (eta + 1.0)))[mask]
        betaq[mask_not] = np.power((1.0 / (2.0 - rand * alpha)), (1.0 / (eta + 1.0)))[mask_not]

        return betaq

    # difference between all variables
    delta = (y2 - y1)

    beta = 1.0 + (2.0 * (y1 - _xl) / delta)
    betaq = calc_betaq(beta)
    c1 = 0.5 * ((y1 + y2) - betaq * delta)

    beta = 1.0 + (2.0 * (_xu - y2) / delta)
    betaq = calc_betaq(beta)
    c2 = 0.5 * ((y1 + y2) + betaq * delta)

    # with the given probability either assign the value from the first or second parent
    b = np.random.random(len(prob_bin)) < prob_bin
    tmp = np.copy(c1[b])
    c1[b] = c2[b]
    c2[b] = tmp

    # first copy the unmodified parents
    Q = np.copy(X)

    # copy the positions where the crossover was done
    Q[0, cross] = c1
    Q[1, cross] = c2

    Q[0] = repair_clamp(Q[0], xl, xu)
    Q[1] = repair_clamp(Q[1], xl, xu)

    return Q


# ---------------------------------------------------------------------------------------------------------
# Class
# ---------------------------------------------------------------------------------------------------------


class SimulatedBinaryCrossover(Crossover):

    def __init__(self,
                 prob_var=0.5,
                 eta=15,
                 prob_exch=0.5,
                 prob_bin=0.5,
                 n_offsprings=2,
                 **kwargs):
        super().__init__(2, n_offsprings, **kwargs)

        self.prob_var = Real(prob_var, bounds=(0.1, 0.9))
        self.eta = Real(eta, bounds=(3.0, 30.0), strict=(1.0, None))
        self.prob_exch = Real(prob_exch, bounds=(0.0, 1.0), strict=(0.0, 1.0))
        self.prob_bin = Real(prob_bin, bounds=(0.0, 1.0), strict=(0.0, 1.0))

    def _do(self, problem, X, **kwargs):
        _, n_matings, _ = X.shape

        # get the parameters required by SBX
        eta, prob_var, prob_exch, prob_bin = get(self.eta, self.prob_var, self.prob_exch, self.prob_bin,
                                                 size=(n_matings, 1))

        # set the binomial probability to zero if no exchange between individuals shall happen
        rand = np.random.random((len(prob_bin), 1))
        prob_bin[rand > prob_exch] = 0.0

        Q = cross_sbx(X.astype(float), problem.xl, problem.xu, eta, prob_var, prob_bin)

        if self.n_offsprings == 1:
            rand = np.random.random(size=n_matings) < 0.5
            Q[0, rand] = Q[1, rand]
            Q = Q[[0]]

        return Q


class SBX(SimulatedBinaryCrossover):
    pass



import numpy as np

from pymoo.visualization.util import plot_axes_arrow, plot_axis_labels, equal_axis, no_ticks, parse_bounds, \
    normalize, get_uniform_points_around_circle
from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot


class StarCoordinate(Plot):

    def __init__(self,
                 axis_extension=1.03,
                 **kwargs):
        """

        Star Coordinate Plot

        Parameters
        ----------------

        axis_style : {axis_style}
        labels : {labels}
        endpoint_style : dict
            Endpoints are drawn at each extreme point of an objective. This style can be modified.

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}

        """
        super().__init__(**kwargs)

        self.axis_extension = axis_extension

        if "arrow_style" not in kwargs:
            self.arrow_style = {
                "head_width": 0.02,
                "head_length": 0.01
            }
        else:
            self.arrow_style = kwargs["arrow_style"]

    def _do(self):

        # initial a figure with a single plot
        self.init_figure()

        # equal axis length and no ticks
        equal_axis(self.ax)
        no_ticks(self.ax)

        # determine the overall scale of points
        _F = np.row_stack([e[0] for e in self.to_plot])
        _min, _max = _F.min(axis=0), _F.max(axis=0)

        V = get_uniform_points_around_circle(self.n_dim)

        plot_axes_arrow(self.ax, V, extend_factor=self.axis_extension, **{**self.axis_style, **self.arrow_style})
        plot_axis_labels(self.ax, V, self.get_labels(), **self.axis_label_style)

        # normalize in range for this plot - here no implicit normalization as in radviz
        bounds = parse_bounds(self.bounds, self.n_dim)
        to_plot_norm = normalize(self.to_plot, bounds)

        for k, (F, kwargs) in enumerate(to_plot_norm):
            N = (F[..., None] * V).sum(axis=1)
            self.ax.scatter(N[:, 0], N[:, 1], **kwargs)


parse_doc_string(StarCoordinate.__init__)



from abc import abstractmethod

import numpy as np

from pymoo.core.operator import Operator


class Selection(Operator):

    def __init__(self, **kwargs) -> None:
        """
        This class is used to select parents for the mating or other evolutionary operators.
        Several strategies can be used to increase the selection pressure.
        """
        super().__init__(**kwargs)

    def do(self, problem, pop, n_select, n_parents, to_pop=True, **kwargs):
        """
        Choose from the population new individuals to be selected.

        Parameters
        ----------


        problem: class
            The problem to be solved. Provides information such as lower and upper bounds or feasibility
            conditions for custom crossovers.

        pop : :class:`~pymoo.core.population.Population`
            The population which should be selected from. Some criteria from the design or objective space
            might be used for the selection. Therefore, only the number of individual might be not enough.

        n_select : int
            Number of individuals to select.

        n_parents : int
            Number of parents needed to create an offspring.

        to_pop : bool
            Whether IF(!) the implementation returns only indices, it should be converted to individuals.

        Returns
        -------
        parents : list
            List of parents to be used in the crossover

        """

        ret = self._do(problem, pop, n_select, n_parents, **kwargs)

        # if some selections return indices they are used to create the individual list
        if to_pop and isinstance(ret, np.ndarray) and ret.dtype == int:
            ret = pop[ret]

        return ret

    @abstractmethod
    def _do(self, problem, pop, n_select, n_parents, **kwargs):
        pass





"""Dynamic Multi-objective Optimization Problems
This Library is a GTA(a and m) benchmarks from the website of the author,
and the implementation of the FDA, DIMP, DMOP, HE benchmarks of the CEC2015 competition
nt: severity of change
taut: frequency of change
tauT: maximum number of generation
tau : current generation

"""
# !/bin/python

import numpy as np
from random import randint
from math import floor, fabs, sin, pi, cos, sqrt

## Parameter configuration ##
LOWER_BOUND = [0.0] + 20 * [-1.0]
UPPER_BOUND = 21 * [1.0]
ERR_MSG = "x is outside decision boundary or dimension of x is not correct"
DELTA_STATE = 1


## Define component functions ##
def beta_uni(x, t, g, obj_num=2):
    """This function is used to calculate the unimodal beta function. Input are
    the decision variable (x), time (t) and g function (g).
    """
    beta = [0.0] * obj_num
    for i in range(obj_num - 1, len(x)):
        beta[(i + 1) % obj_num] += (x[i] - g(x, t)) * (x[i] - g(x, t))

    beta = [(2.0 / int(len(LOWER_BOUND) / obj_num)) * b for b in beta]
    return beta


def beta_multi(x, t, g, obj_num=2):
    """This function is used to calculate the multi-modal beta function. Input
    are the decision variable (x), time (t) and g function (g).
    """
    beta = [0.0] * obj_num
    for i in range(obj_num - 1, len(x)):
        beta[(i + 1) % obj_num] += (x[i] - g(x, t)) * (x[i] - g(x, t)) * \
                                   (1 + np.abs(np.sin(4 * np.pi * (x[i] - g(x, t)))))

    beta = [(2.0 / int(len(LOWER_BOUND) / obj_num)) * b for b in beta]
    return beta


def beta_mix(x, t, g, obj_num=2):
    """This function is used to calculate the mixed unimodal and multi-modal
    beta function. Input are the decision variable (x), time (t) and g function
    (g).
    """
    beta = [0.0] * obj_num
    k = int(abs(5.0 * (int(DELTA_STATE * int(t) / 5.0) % 2) - (DELTA_STATE * int(t) % 5)))

    for i in range(obj_num - 1, len(x)):
        temp = 1.0 + (x[i] - g(x, t)) * (x[i] - g(x, t)) - np.cos(2 * np.pi * k * (x[i] - g(x, t)))
        beta[(i + 1) % obj_num] += temp
    beta = [(2.0 / int(len(LOWER_BOUND) / obj_num)) * b for b in beta]
    return beta


def alpha_conv(x):
    """This function is used to calculate the alpha function with convex POF.
    Input is decision variable (x).
    """
    return [x[0], 1 - np.sqrt(x[0])]


def alpha_disc(x):
    """This function is used to calculate the alpha function with discrete POF.
    Input is decision variable (x).
    """
    return [x[0], 1.5 - np.sqrt(x[0]) - 0.5 * np.sin(10 * np.pi * x[0])]


def alpha_mix(x, t):
    """This function is used to calculate the alpha function with mixed
    continuous POF and discrete POF.
    """
    k = int(abs(5.0 * (int(DELTA_STATE * int(t) / 5.0) % 2) - (DELTA_STATE * int(t) % 5)))
    return [x[0], 1 - np.sqrt(x[0]) + 0.1 * k * (1 + np.sin(10 * np.pi * x[0]))]


def alpha_conf(x, t):
    """This function is used to calculate the alpha function with time-varying
    conflicting objective. Input are decision variable (x) and time (t).
    """
    k = int(abs(5.0 * (int(DELTA_STATE * int(t) / 5.0) % 2) - (DELTA_STATE * int(t) % 5)))
    return [x[0], 1 - np.power(x[0], \
                               np.log(1 - 0.1 * k) / np.log(0.1 * k + np.finfo(float).eps))]


def alpha_conf_3obj_type1(x, t):
    """This function is used to calculate the alpha unction with time-varying
    conflicting objective (3-objective, type 1). Input are decision variable
    (x) and time (t).
    """
    k = int(abs(5.0 * (int(DELTA_STATE * int(t) / 5.0) % 2) - (DELTA_STATE * int(t) % 5)))
    alpha1 = fix_numerical_instability(np.cos(0.5 * x[0] * np.pi) * np.cos(0.5 * x[1] * np.pi))
    alpha2 = fix_numerical_instability(np.cos(0.5 * x[0] * np.pi) * np.sin(0.5 * x[1] * np.pi))
    alpha3 = fix_numerical_instability(np.sin(0.5 * x[0] * np.pi + 0.25 * (k / 5.0) * np.pi))
    return [alpha1, alpha2, alpha3]


def alpha_conf_3obj_type2(x, t):
    """This function is used to calculate the alpha unction with time-varying
    conflicting objective (3-objective, type 2). Input are decision variable (x)
    and time (t).
    """
    k = int(abs(5.0 * (int(DELTA_STATE * int(t) / 5.0) % 2) - (DELTA_STATE * int(t) % 5)))
    k_ratio = (5.0 - k) / 5.0
    alpha1 = fix_numerical_instability(np.cos(0.5 * x[0] * np.pi) * np.cos(0.5 * x[1] * np.pi * k_ratio))
    alpha2 = fix_numerical_instability(np.cos(0.5 * x[0] * np.pi) * np.sin(0.5 * x[1] * np.pi * k_ratio))
    alpha3 = fix_numerical_instability(np.sin(0.5 * x[0] * np.pi))
    return [alpha1, alpha2, alpha3]


def g(x, t):
    """This function is used to calculate the g function used in the paper.
    Input are decision variable (x) and time (t).
    """
    return np.sin(0.5 * np.pi * (t - x[0]))


## Utility functions ##
def check_boundary(x, upper_bound=UPPER_BOUND, lower_bound=LOWER_BOUND):
    """Check the dimension of x and whether it is in the decision boundary. x is
    decision variable, upper_bound and lower_bound are upperbound and lowerbound
    lists of the decision space
    """
    if len(x) != len(upper_bound) or len(x) != len(lower_bound):
        return False

    output = True
    for e, upp, low in zip(x, upper_bound, lower_bound):
        output = output and (e >= low) and (e <= upp)
    return output


def check_boundary_3obj(x, upper_bound=UPPER_BOUND, lower_bound=LOWER_BOUND):
    """Check the dimension of x and whether it is in the decision boundary. x is
    decision variable, upper_bound and lower_bound are upperbound and lowerbound
    lists of the decision space
    """
    lower_bound = [0.0] + lower_bound
    upper_bound = [1.0] + upper_bound
    if len(x) != len(upper_bound) or len(x) != len(lower_bound):
        return False

    output = True
    for e, upp, low in zip(x, upper_bound, lower_bound):
        output = output and (e >= low) and (e <= upp)
    return output


def fix_numerical_instability(x):
    """Check whether x is close to zero, sqrt(0.5) or not. If it is close to
    these two values, changes x to the value. Otherwise, return x.
    """
    if np.allclose(0.0, x):
        return 0.0

    if np.allclose(np.sqrt(0.5), x):
        return np.sqrt(0.5)
    return x


def additive(alpha, beta):
    """Additive form of the benchmark problem.
    """
    return [a + b for a, b in zip(alpha, beta)]


#    return [alpha[0] + beta[0], alpha[1] + beta[1]]


def multiplicative(alpha, beta):
    """Multiplicative form of the benchmark problem.
    """
    return [a * (1 + b) for a, b in zip(alpha, beta)]


#    return [alpha[0]*(1 + beta[0]), alpha[1]*(1 + beta[1])]


## Benchmark functions ##
def DB1a(x, t):
    """DB1a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conv(x)
        beta = beta_uni(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB1m(x, t):
    """DB1m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conv(x)
        beta = beta_uni(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB2a(x, t):
    """DB2a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conv(x)
        beta = beta_multi(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB2m(x, t):
    """DB2m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conv(x)
        beta = beta_multi(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB3a(x, t):
    """DB3a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conv(x)
        beta = beta_mix(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB3m(x, t):
    """DB3m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conv(x)
        beta = beta_mix(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB4a(x, t):
    """DB4a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_disc(x)
        beta = beta_mix(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB4m(x, t):
    """DB4m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_disc(x)
        beta = beta_mix(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB5a(x, t):
    """DB5a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_mix(x, t)
        beta = beta_multi(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB5m(x, t):
    """DB5m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_mix(x, t)
        beta = beta_multi(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB6a(x, t):
    """DB6a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_mix(x, t)
        beta = beta_mix(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB6m(x, t):
    """DB6m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_mix(x, t)
        beta = beta_mix(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB7a(x, t):
    """DB7a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf(x, t)
        beta = beta_multi(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB7m(x, t):
    """DB7m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf(x, t)
        beta = beta_multi(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB8a(x, t):
    """DB8a dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf(x, t)
        beta = beta_mix(x, t, g)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB8m(x, t):
    """DB8m dynamic benchmark problem
    """
    if check_boundary(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf(x, t)
        beta = beta_mix(x, t, g)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB9a(x, t):
    """DB9a dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type1(x, t)
        beta = beta_multi(x, t, g, obj_num=3)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB9m(x, t):
    """DB9m dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type1(x, t)
        beta = beta_multi(x, t, g, obj_num=3)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB10a(x, t):
    """DB10a dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type1(x, t)
        beta = beta_mix(x, t, g, obj_num=3)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB10m(x, t):
    """DB10m dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type1(x, t)
        beta = beta_mix(x, t, g, obj_num=3)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB11a(x, t):
    """DB11a dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type2(x, t)
        beta = beta_multi(x, t, g, obj_num=3)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB11m(x, t):
    """DB11m dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type2(x, t)
        beta = beta_multi(x, t, g, obj_num=3)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB12a(x, t):
    """DB12a dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type2(x, t)
        beta = beta_mix(x, t, g, obj_num=3)
        return additive(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def DB12m(x, t):
    """DB12m dynamic benchmark problem
    """
    if check_boundary_3obj(x, UPPER_BOUND, LOWER_BOUND):
        alpha = alpha_conf_3obj_type2(x, t)
        beta = beta_mix(x, t, g, obj_num=3)
        return multiplicative(alpha, beta)
    else:
        raise Exception(ERR_MSG)


def fda2_deb(x, t):
    f1 = x[0]
    H = 2 * np.sin(0.5 * np.pi * (t - 1))
    XII = x[1:6]
    XIII = x[6:13]
    g = 1 + np.sum(np.power(XII, 2))
    Htemp = np.sum(np.power((XIII - H / 4), 2))
    h = 1 - np.power((f1 / g), np.power(2, H + Htemp))
    f2 = g * h
    return [f1, f2]


def FDA4(X, t):
    """FDA4 dynamic benchmark problem
    """
    XII = X[2:]
    G = fabs(sin(0.5 * pi * t))
    g = sum([pow(xi - G, 2) for xi in XII])
    f1 = (1 + g) * cos(X[0] * pi / 2) * cos(X[1] * pi / 2)
    f2 = (1 + g) * cos(X[0] * pi / 2) * sin(X[1] * pi / 2)
    f3 = (1 + g) * sin(X[0] * pi / 2)
    return [f1, f2, f3]


def FDA5(X, t):
    """FDA5 dynamic benchmark problem
    """
    XII = X[2:]
    G = fabs(sin(0.5 * pi * t))
    g = G + sum([pow(xi - G, 2) for xi in XII])
    F = 1 + 100 * pow(sin(0.5 * pi * t), 4)
    y = lambda i: pow(X[i], F)
    f1 = (1 + g) * cos(y(0) * pi / 2) * cos(y(1) * pi / 2)
    f2 = (1 + g) * cos(y(0) * pi / 2) * sin(y(1) * pi / 2)
    f3 = (1 + g) * sin(y(0) * pi / 2)
    return [f1, f2, f3]


def DIMP2(X, t):
    """DIMP2 dynamic benchmark problem
    """
    n = len(X)
    XII = X[1:]
    g = 1.0 + 2.0 * (len(XII))
    for k in range(1, n):
        G = sin(pow(0.5 * pi * t + 2.0 * pi * float(k + 1) / float(n + 1.0), 2))
        g += pow(X[k] - G, 2) - 2.0 * cos(3.0 * pi * (X[k] - G))
    f1 = X[0]
    h = 1 - sqrt(f1 / g)
    f2 = g * h
    return [f1, f2]


def dMOP2(X, t):
    """dMOP2 dynamic benchmark problem
    """
    XII = X[1:]
    G = sin(0.5 * pi * t)
    g = 1 + 9 * sum([pow(xi - G, 2) for xi in XII])
    H = 0.75 * sin(0.5 * pi * t) + 1.25
    f1 = X[0]
    h = 1 - pow((f1 / g), H)
    f2 = g * h
    return [f1, f2]


def dMOP3(X, tau, nt, taut, r, rIteration):
    """dMOP3 dynamic benchmark problem
    """
    if tau % taut == 0 and rIteration != tau:
        r = randint(0, 9)
        rIteration = tau

    XII = X[:r] + X[r + 1:]
    t = float(1) / float(nt)
    t = t * floor(float(tau) / float(taut))
    G = sin(0.5 * pi * t)
    g = 1 + 9 * sum([pow(xi - G, 2) for xi in XII])
    H = 0.75 * sin(0.5 * pi * t)
    f1 = X[r]
    f2 = 1 - pow((f1 / g), H)
    return [f1, f2, r, rIteration]


def HE2(X, t):
    """HE2 dynamic benchmark problem
    """
    n = 30
    XII = X[1:]
    H = 0.75 * sin(0.5 * pi * t) + 1.25
    g = 1 + (9 / (n - 1)) * sum(XII)
    f1 = X[0]
    h = 1 - pow(sqrt(f1 / g), H) - pow(f1 / g, H) * sin(10 * pi * f1)
    f2 = g * h
    return [f1, f2]


def HE7(X, t):
    """HE7 dynamic benchmark problem
    """

    def _f1(input1):
        value = input1[0]
        ssum = 0.0
        index = 0
        for k in range(2, len(input1), 2):
            val = 6 * pi * value + k * pi / len(input1)
            ssum += pow(input1[k] - (0.3 * value * value * cos(4 * val) + 0.6 * value) * cos(val), 2)
            index += 1
        ssum *= 2.0 / index
        ssum += value
        return ssum

    def _g(input1):
        value = input1[0]
        ssum = 0.0
        index = 0
        for k in range(1, len(input1), 2):
            val = 6 * pi * value + k * pi / len(input1)
            ssum += pow(input1[k] - (0.3 * value * value * cos(4 * val) + 0.6 * value) * sin(val), 2)
            index += 1
        ssum *= 2.0 / index
        ssum += 2.0 - sqrt(value)
        return ssum

    f1 = _f1(X)
    g = _g(X)
    H = 0.75 * sin(0.5 * pi * t) + 1.25
    h = 1 - pow(f1 / g, H)
    f2 = g * h
    return [f1, f2]


def HE9(X, t):
    """HE9 dynamic benchmark problem
    """

    def _f1(input1):
        value = input1[0]
        ssum = 0.0
        index = 0
        for k in range(2, len(input1), 2):
            ssum += pow(input1[k] - sin(6 * pi * value + k * pi / len(input1)), 2)
            index += 1
        ssum *= 2.0 / index
        ssum += value
        return ssum

    def _g(input1):
        value = input1[0]
        ssum = 0.0
        index = 0
        for k in range(1, len(input1), 2):
            ssum += pow(input1[k] - sin(6 * pi * value + k * pi / len(input1)), 2)
            index += 1
        ssum *= 2.0 / index
        ssum += 2.0 - pow(value, 2)
        return ssum

    f1 = _f1(X)
    g = _g(X)
    H = 0.75 * sin(0.5 * pi * t) + 1.25
    h = 1 - pow(f1 / g, H)
    f2 = g * h
    return [f1, f2]



import numpy as np

from pymoo.core.callback import Callback
from pymoo.indicators.igd import IGD


class MIGD(Callback):

    def __init__(self, reevaluate=True) -> None:
        """
        Mean Inverted Generational Distance (MIGD)

        For dynamic optimization problems the performance metric needs to involve the IGD value over time as the
        problem is changing. Thus, the performance needs to be evaluated in each iteration for which
        defining a callback is ideal.

        """

        super().__init__()

        # whether the MIGD should be based on reevaluated solutions
        self.reevaluate = reevaluate

        # the list where each of the recordings are stored: timesteps and igd
        self.records = []

    def update(self, algorithm, **kwargs):

        # the problem to be solved
        problem = algorithm.problem
        assert problem.n_constr == 0, "The current implementation only works for unconstrained problems!"

        # the current time
        t = problem.time

        # the current pareto-front of the problem (at the specific time step)
        pf = problem.pareto_front()

        # the current population of the algorithm
        pop = algorithm.pop

        # if the callback should reevaluate to match the current time step and avoid deprecated values
        if self.reevaluate:
            X = pop.get("X")
            F = problem.evaluate(X, return_values_of=["F"])
        else:
            F = pop.get("F")

        # calculate the current igd values
        igd = IGD(pf).do(F)

        self.records.append((t, igd))

    def value(self):
        return np.array([igd for _, igd in self.records]).mean()



import os

import numpy as np

from pymoo.core.problem import Problem
from pymoo.util.remote import Remote


class COCOProblem(Problem):

    def __init__(self, name, n_var=10, pf_from_file=True, **kwargs):
        self.function, self.instance, self.object = get_bbob(name, n_var)
        self.name = name
        self.pf_from_file = pf_from_file

        coco = self.object
        n_var, n_obj, n_ieq_constr = coco.number_of_variables, coco.number_of_objectives, coco.number_of_constraints
        xl, xu = coco.lower_bounds, coco.upper_bounds

        super().__init__(n_var=n_var,
                         n_obj=n_obj,
                         n_ieq_constr=n_ieq_constr,
                         xl=xl,
                         xu=xu,
                         **kwargs)

    def _calc_pareto_set(self, *args, **kwargs):
        if self.n_obj == 1:
            fname = '._bbob_problem_best_parameter.txt'

            self.object._best_parameter(what="print")
            ps = np.loadtxt(fname)
            os.remove(fname)

            return ps

    def _calc_pareto_front(self, *args, **kwargs):
        if self.pf_from_file:
            return Remote.get_instance().load("pymoo", "pf", "bbob.pf", to="json")[str(self.function)][str(self.instance)]
        else:
            ps = self.pareto_set()
            if ps is not None:
                return self.evaluate(ps)

    def _evaluate(self, X, out, *args, **kwargs):
        out["F"] = np.array([self.object(x) for x in X])

    def __getstate__(self):
        d = self.__dict__.copy()
        d["object"] = None
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.object = get_bbob(self.name, self.n_var)


def get_bbob(name, n_var=10, **kwargs):
    try:
        import cocoex as ex
    except:
        raise Exception("COCO test suite not found. \nInstallation Guide: https://github.com/numbbo/coco")

    args = name.split("-")

    n_instance = int(args[-1])
    n_function = int(args[-2].replace("f", ""))

    assert 1 <= n_function <= 24, f"BBOB has 24 different functions to be chosen. {n_function} is out of range."

    suite_filter_options = f"function_indices: {n_function} " \
                           f"instance_indices: {n_instance} " \
                           f"dimensions: {n_var}"

    problems = ex.Suite("bbob", "", suite_filter_options)
    assert len(problems) == 1, "COCO problem not found."

    coco = problems.next_problem()

    return n_function, n_instance, coco





from pymoo.core.termination import Termination


class TerminationCollection(Termination):

    def __init__(self, *args) -> None:
        super().__init__()
        self.terminations = args

    def _update(self, algorithm):
        return max([termination.update(algorithm) for termination in self.terminations])



import numpy as np

from pymoo.util.dominator import Dominator


def fast_non_dominated_sort(F, **kwargs):
    M = Dominator.calc_domination_matrix(F)

    # calculate the dominance matrix
    n = M.shape[0]

    fronts = []

    if n == 0:
        return fronts

    # final rank that will be returned
    n_ranked = 0
    ranked = np.zeros(n, dtype=int)

    # for each individual a list of all individuals that are dominated by this one
    is_dominating = [[] for _ in range(n)]

    # storage for the number of solutions dominated this one
    n_dominated = np.zeros(n)

    current_front = []

    for i in range(n):

        for j in range(i + 1, n):
            rel = M[i, j]
            if rel == 1:
                is_dominating[i].append(j)
                n_dominated[j] += 1
            elif rel == -1:
                is_dominating[j].append(i)
                n_dominated[i] += 1

        if n_dominated[i] == 0:
            current_front.append(i)
            ranked[i] = 1.0
            n_ranked += 1

    # append the first front to the current front
    fronts.append(current_front)

    # while not all solutions are assigned to a pareto front
    while n_ranked < n:

        next_front = []

        # for each individual in the current front
        for i in current_front:

            # all solutions that are dominated by this individuals
            for j in is_dominating[i]:
                n_dominated[j] -= 1
                if n_dominated[j] == 0:
                    next_front.append(j)
                    ranked[j] = 1.0
                    n_ranked += 1

        fronts.append(next_front)
        current_front = next_front

    return fronts




from pymoo.core.termination import Termination


class MinimumFunctionValueTermination(Termination):

    def __init__(self, fmin) -> None:
        super().__init__()
        self.fmin = fmin

    def _update(self, algorithm):
        opt = algorithm.opt

        if not any(opt.get("feas")):
            return 0.0
        else:
            return self.fmin / opt.get("F").min()



import numpy as np

from pymoo.core.problem import Problem
# Based on the C++ implementation by the Ma and Wang
# http://www.escience.cn/people/yongwang1/index.html
from pymoo.problems.many import get_ref_dirs
from pymoo.util.remote import Remote


class MW(Problem):
    def __init__(self, n_var, n_obj, n_ieq_constr, **kwargs):
        if 'xl' not in kwargs:
            kwargs['xl'] = 0
        if 'xu' not in kwargs:
            kwargs['xu'] = 1
        super().__init__(n_var=n_var,
                         n_obj=n_obj,
                         n_ieq_constr=n_ieq_constr,
                         vtype=float, **kwargs)

    @staticmethod
    def LA1(A, B, C, D, theta):
        return A * np.power(np.sin(B * np.pi * np.power(theta, C)), D)

    @staticmethod
    def LA2(A, B, C, D, theta):
        return A * np.power(np.sin(B * np.power(theta, C)), D)

    @staticmethod
    def LA3(A, B, C, D, theta):
        return A * np.power(np.cos(B * np.power(theta, C)), D)

    def g1(self, X):
        d = self.n_var
        n = d - self.n_obj

        z = np.power(X[:, self.n_obj - 1:], n)
        i = np.arange(self.n_obj - 1, d)

        exp = 1 - np.exp(-10.0 * (z - 0.5 - i / (2 * d)) * (z - 0.5 - i / (2 * d)))
        distance = 1 + exp.sum(axis=1)
        return distance

    def g2(self, X):
        d = self.n_var
        n = d

        i = np.arange(self.n_obj - 1, d)
        z = 1 - np.exp(-10.0 * (X[:, self.n_obj - 1:] - i / n) * (X[:, self.n_obj - 1:] - i / n))
        contrib = (0.1 / (n)) * z * z + 1.5 - 1.5 * np.cos(2 * np.pi * z)
        distance = 1 + contrib.sum(axis=1)
        return distance

    def g3(self, X):
        contrib = 2.0 * np.power(
            X[:, self.n_obj - 1:] + (X[:, self.n_obj - 2:-1] - 0.5) * (X[:, self.n_obj - 2:-1] - 0.5) - 1.0, 2.0)
        distance = 1 + contrib.sum(axis=1)
        return distance


class MW1(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 1)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)
        f0 = X[:, 0]
        f1 = g * (1 - 0.85 * f0 / g)

        g0 = f0 + f1 - 1 - self.LA1(0.5, 2.0, 1.0, 8.0, np.sqrt(2.0) * f1 - np.sqrt(2.0) * f0)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            F = np.zeros((100, 2))
            F[:, 0] = np.linspace(0, 1, 100)
        else:
            F = ref_dirs
        F[:, 1] = 1 - 0.85 * F[:, 0]
        l = np.sqrt(2) * F[:, 1] - np.sqrt(2) * F[:, 0]
        c = 1 - F[:, 0] - F[:, 1] + 0.5 * np.sin(2 * np.pi * l) ** 8
        F = F[c >= 0]
        return F


class MW2(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 1)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)
        f0 = X[:, 0]
        f1 = g * (1 - f0 / g)

        g0 = f0 + f1 - 1 - self.LA1(0.5, 3.0, 1.0, 8.0, np.sqrt(2.0) * f1 - np.sqrt(2.0) * f0)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            F = np.zeros((100, 2))
            F[:, 0] = np.linspace(0, 1, 100)
        else:
            F = ref_dirs
        F[:, 1] = 1 - F[:, 0]
        return F


class MW3(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 2)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g3(X)
        f0 = X[:, 0]
        f1 = g * (1 - f0 / g)

        g0 = f0 + f1 - 1.05 - self.LA1(0.45, 0.75, 1.0, 6.0, np.sqrt(2.0) * f1 - np.sqrt(2.0) * f0)
        g1 = 0.85 - f0 - f1 + self.LA1(0.3, 0.75, 1.0, 2.0, np.sqrt(2.0) * f1 - np.sqrt(2.0) * f0)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1])

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            F = np.zeros((100, 2))
            F[:, 0] = np.linspace(0, 1, 100)
        else:
            F = ref_dirs
        F[:, 1] = 1 - F[:, 0]
        invalid = (0.85 - F[:, 0] - F[:, 1] + 0.3 * np.sin(0.75 * np.pi * np.sqrt(2) * (F[:, 1] - F[:, 0])) ** 2) > 0
        while invalid.any():
            F[invalid, :] *= 1.001
            invalid = (0.85 - F[:, 0] - F[:, 1] + 0.3 * np.sin(
                0.75 * np.pi * np.sqrt(2) * (F[:, 1] - F[:, 0])) ** 2) > 0
        return F


class MW4(MW):
    def __init__(self, n_var=None, n_obj=3, **kwargs):
        if n_var is None:
            n_var = n_obj + 12
        super().__init__(n_var, n_obj, 1)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)
        f = g.reshape((-1, 1)) * np.ones((X.shape[0], self.n_obj))
        f[:, 1:] *= X[:, (self.n_obj - 2)::-1]
        f[:, 0:-1] *= np.flip(np.cumprod(1 - X[:, :(self.n_obj - 1)], axis=1), axis=1)

        g0 = f.sum(axis=1) - 1 - self.LA1(0.4, 2.5, 1.0, 8.0, f[:, -1] - f[:, :-1].sum(axis=1))
        out["F"] = f
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        F = ref_dirs
        l = F[:, -1] - np.sum(F[:, :-1], axis=1)
        c = (1 + 0.4 * np.sin(2.5 * np.pi * l) ** 8) - np.sum(F, axis=1)
        return F[c >= 0]


class MW5(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 3)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)
        f0 = g * X[:, 0]
        f1 = g * np.sqrt(1.0 - np.power(f0 / g, 2.0))

        with np.errstate(divide='ignore'):
            atan = np.arctan(f1 / f0)

        g0 = f0 ** 2 + f1 ** 2 - np.power(1.7 - self.LA2(0.2, 2.0, 1.0, 1.0, atan), 2.0)
        t = 0.5 * np.pi - 2 * np.abs(atan - 0.25 * np.pi)
        g1 = np.power(1 + self.LA2(0.5, 6.0, 3.0, 1.0, t), 2.0) - f0 ** 2 - f1 ** 2
        g2 = np.power(1 - self.LA2(0.45, 6.0, 3.0, 1.0, t), 2.0) - f0 ** 2 - f1 ** 2
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1, g2])

    def _calc_pareto_front(self, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "MW", "MW5.pf")


class MW6(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 1, xl=0.0, xu=1.1)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)
        f0 = g * X[:, 0]
        f1 = g * np.sqrt(1.1 * 1.1 - np.power(f0 / g, 2.0))

        with np.errstate(divide='ignore'):
            atan = np.arctan(f1 / f0)

        g0 = f0 ** 2 / np.power(1.0 + self.LA3(0.15, 6.0, 4.0, 10.0, atan), 2.0) + f1 ** 2 / np.power(
            1.0 + self.LA3(0.75, 6.0, 4.0, 10.0, atan), 2.0) - 1
        out["F"] = np.column_stack([f0, f1])
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            F = np.zeros((100, 2))
            F[:, 0] = np.linspace(0, 1, 100)
        else:
            F = ref_dirs
        F[:, 1] = 1 - F[:, 0]
        F = F / np.sqrt(np.sum(F ** 2, axis=1) / 1.21).reshape((-1, 1))
        l = np.cos(6 * np.arctan(F[:, 1] / F[:, 0]) ** 4) ** 10
        c = 1 - (F[:, 0] / (1 + 0.15 * l)) ** 2 - (F[:, 1] / (1 + 0.75 * l)) ** 2
        return F[c >= 0]


class MW7(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 2)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g3(X)
        f0 = g * X[:, 0]
        f1 = g * np.sqrt(1 - np.power(f0 / g, 2))

        with np.errstate(divide='ignore'):
            atan = np.arctan(f1 / f0)

        g0 = f0 ** 2 + f1 ** 2 - np.power(1.2 + np.abs(self.LA2(0.4, 4.0, 1.0, 16.0, atan)), 2.0)
        g1 = np.power(1.15 - self.LA2(0.2, 4.0, 1.0, 8.0, atan), 2.0) - f0 ** 2 - f1 ** 2
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1])

    def _calc_pareto_front(self, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "MW", "MW7.pf")


class MW8(MW):
    def __init__(self, n_var=None, n_obj=3, **kwargs):
        if n_var is None:
            n_var = n_obj + 12
        super().__init__(n_var, n_obj, 1)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)
        f = g.reshape((-1, 1)) * np.ones((X.shape[0], self.n_obj))
        f[:, 1:] *= np.sin(0.5 * np.pi * X[:, (self.n_obj - 2)::-1])
        cos = np.cos(0.5 * np.pi * X[:, :(self.n_obj - 1)])
        f[:, 0:-1] *= np.flip(np.cumprod(cos, axis=1), axis=1)

        f_squared = (f ** 2).sum(axis=1)
        g0 = f_squared - (1.25 - self.LA2(0.5, 6.0, 1.0, 2.0, np.arcsin(f[:, -1] / np.sqrt(f_squared)))) * (
                1.25 - self.LA2(0.5, 6.0, 1.0, 2.0, np.arcsin(f[:, -1] / np.sqrt(f_squared))))
        out["F"] = f
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)

        F = ref_dirs
        F = F / np.sqrt(np.sum(F ** 2, axis=1)).reshape((-1, 1))
        c = (1.25 - 0.5 * np.sin(6 * np.arcsin(F[:, -1])) ** 2) ** 2 - np.sum(F ** 2, axis=1)
        return F[c >= 0]


class MW9(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 1)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)
        f0 = g * X[:, 0]
        f1 = g * (1.0 - np.power(f0 / g, 0.6))

        t1 = (1 - 0.64 * f0 * f0 - f1) * (1 - 0.36 * f0 * f0 - f1)
        t2 = (1.35 * 1.35 - (f0 + 0.35) * (f0 + 0.35) - f1) * (1.15 * 1.15 - (f0 + 0.15) * (f0 + 0.15) - f1)
        g0 = np.minimum(t1, t2)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "MW", "MW9.pf")


class MW10(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 3)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)
        f0 = g * np.power(X[:, 0], self.n_var)
        f1 = g * (1.0 - np.power(f0 / g, 2.0))

        g0 = -1.0 * (2.0 - 4.0 * f0 * f0 - f1) * (2.0 - 8.0 * f0 * f0 - f1)
        g1 = (2.0 - 2.0 * f0 * f0 - f1) * (2.0 - 16.0 * f0 * f0 - f1)
        g2 = (1.0 - f0 * f0 - f1) * (1.2 - 1.2 * f0 * f0 - f1)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1, g2])

    def _calc_pareto_front(self, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "MW", "MW10.pf")


class MW11(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 4, xl=0.0, xu=np.sqrt(2))

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g3(X)
        f0 = g * X[:, 0]
        f1 = g * np.sqrt(2.0 - np.power(f0 / g, 2.0))

        g0 = -1.0 * (3.0 - f0 * f0 - f1) * (3.0 - 2.0 * f0 * f0 - f1)
        g1 = (3.0 - 0.625 * f0 * f0 - f1) * (3.0 - 7.0 * f0 * f0 - f1)
        g2 = -1.0 * (1.62 - 0.18 * f0 * f0 - f1) * (1.125 - 0.125 * f0 * f0 - f1)
        g3 = (2.07 - 0.23 * f0 * f0 - f1) * (0.63 - 0.07 * f0 * f0 - f1)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1, g2, g3])

    def _calc_pareto_front(self, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "MW", "MW11.pf")


class MW12(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 2)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)
        f0 = g * X[:, 0]
        f1 = g * (0.85 - 0.8 * (f0 / g) - 0.08 * np.abs(np.sin(3.2 * np.pi * (f0 / g))))

        g0 = -1.0 * (1 - 0.625 * f0 - f1 + 0.08 * np.sin(2 * np.pi * (f1 - f0 / 1.6))) * (
                1.4 - 0.875 * f0 - f1 + 0.08 * np.sin(2 * np.pi * (f1 / 1.4 - f0 / 1.6)))
        g1 = (1 - 0.8 * f0 - f1 + 0.08 * np.sin(2 * np.pi * (f1 - f0 / 1.5))) * (
                1.8 - 1.125 * f0 - f1 + 0.08 * np.sin(2 * np.pi * (f1 / 1.8 - f0 / 1.6)))
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1])

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            F = np.zeros((100, 2))
            F[:, 0] = np.linspace(0, 1, 100)
        else:
            F = ref_dirs
        F[:, 1] = 0.85 - 0.8 * F[:, 0] - 0.08 * np.abs(np.sin(3.2 * np.pi * F[:, 0]))

        invalid = (1 - 0.8 * F[:, 0] - F[:, 1] + 0.08 * np.sin(2 * np.pi * (F[:, 1] - F[:, 0] / 1.5))) * (
                1.8 - 1.125 * F[:, 0] - F[:, 1] + 0.08 * np.sin(2 * np.pi * (F[:, 1] / 1.8 - F[:, 0] / 1.6))) > 0
        while invalid.any():
            F[invalid, :] *= 1.001
            invalid = (1 - 0.8 * F[:, 0] - F[:, 1] + 0.08 * np.sin(2 * np.pi * (F[:, 1] - F[:, 0] / 1.5))) * (
                    1.8 - 1.125 * F[:, 0] - F[:, 1] + 0.08 * np.sin(
                2 * np.pi * (F[:, 1] / 1.8 - F[:, 0] / 1.6))) > 0
        return F


class MW13(MW):
    def __init__(self, n_var=15, **kwargs):
        super().__init__(n_var, 2, 2, xu=1.5)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)
        f0 = g * X[:, 0]
        f1 = g * (5.0 - np.exp(f0 / g) - np.abs(0.5 * np.sin(3 * np.pi * f0 / g)))

        g0 = -1.0 * (5.0 - (1 + f0 + 0.5 * f0 * f0) - 0.5 * np.sin(3 * np.pi * f0) - f1) * (
                5.0 - (1 + 0.7 * f0) - 0.5 * np.sin(3 * np.pi * f0) - f1)
        g1 = (5.0 - np.exp(f0) - 0.5 * np.sin(3 * np.pi * f0) - f1) * (
                5.0 - (1 + 0.4 * f0) - 0.5 * np.sin(3 * np.pi * f0) - f1)
        out["F"] = np.column_stack([f0, f1])
        out["G"] = np.column_stack([g0, g1])

    def _calc_pareto_front(self, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "MW", "MW13.pf")


class MW14(MW):
    def __init__(self, n_var=None, n_obj=3, **kwargs):
        if n_var is None:
            n_var = n_obj + 12
        super().__init__(n_var, n_obj, 1, xu=1.5)

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g3(X)
        f = np.zeros((X.shape[0], self.n_obj))
        f[:, :-1] = X[:, :(self.n_obj - 1)]
        LA1 = self.LA1(1.5, 1.1, 2.0, 1.0, f[:, :-1])
        inter = (6 - np.exp(f[:, :-1]) - LA1).sum(axis=1)
        f[:, -1] = g / (self.n_obj - 1) * inter

        alpha = 6.1 - 1 - f[:, :-1] - 0.5 * f[:, :-1] * f[:, :-1] - LA1
        g0 = f[:, -1] - 1 / (self.n_obj - 1) * alpha.sum(axis=1)
        out["F"] = f
        out["G"] = g0.reshape((-1, 1))

    def _calc_pareto_front(self, **kwargs):
        if self.n_obj == 3:
            return Remote.get_instance().load("pymoo", "pf", "MW", "MW14.pf")



import abc

from pymoo.util.normalization import PreNormalization


class Indicator(PreNormalization):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # what should an indicator return if no solutions are provided is defined here
        self.default_if_empty = 0.0

    def __call__(self, F, *args, **kwargs):
        return self.do(F, *args, **kwargs)

    def do(self, F, *args, **kwargs):

        # if it is a 1d array
        if F.ndim == 1:
            F = F[None, :]

        # if no points have been provided just return the default
        if len(F) == 0:
            return self.default_if_empty

        # do the normalization - will only be done if zero_to_one is enabled
        F = self.normalization.forward(F)

        return self._do(F, *args, **kwargs)

    @abc.abstractmethod
    def _do(self, F, *args, **kwargs):
        return



from pymoo.core.callback import Callback
from pymoo.util.display.progress import ProgressBar


class Display(Callback):

    def __init__(self, output=None, progress=False, verbose=False):
        super().__init__()
        self.output = output
        self.verbose = verbose
        self.progress = ProgressBar() if progress else None

    def update(self, algorithm, **kwargs):
        output, progress = self.output, self.progress

        if self.verbose and output:
            text = ""
            header = not output.is_initialized
            output(algorithm)

            if header:
                text += output.header(border=True) + '\n'
            text += output.text()

            print(text)

        if progress:
            perc = algorithm.termination.perc
            progress.set(perc)

    def finalize(self):

        if self.progress:
            self.progress.close()



from pymoo.core.meta import Meta
from pymoo.core.problem import Problem


class MultiToSingleObjective(Meta, Problem):

    def __init__(self, problem, decomposition, kwargs=None):
        super().__init__(problem)
        self.decomposition = decomposition
        self.kwargs = kwargs if not None else dict()
        self.n_obj = 1

    def do(self, X, return_values_of, *args, **kwargs):
        out = self.__object__.do(X, return_values_of, *args, **kwargs)
        F = out["F"]
        out["__F__"] = F
        out["F"] = self.decomposition.do(F, **self.kwargs)[:, None]
        return out







import numpy as np

from pymoo.core.mutation import Mutation
from pymoo.core.variable import Real, get
from pymoo.operators.repair.bounds_repair import repair_random_init


# ---------------------------------------------------------------------------------------------------------
# Function
# ---------------------------------------------------------------------------------------------------------


def mut_gauss(X, xl, xu, sigma, prob):
    n, n_var = X.shape
    assert len(sigma) == n
    assert len(prob) == n

    Xp = np.full(X.shape, np.inf)

    mut = np.random.random(X.shape) < prob[:, None]

    Xp[:, :] = X

    _xl = np.repeat(xl[None, :], X.shape[0], axis=0)[mut]
    _xu = np.repeat(xu[None, :], X.shape[0], axis=0)[mut]
    sigma = sigma[:, None].repeat(n_var, axis=1)[mut]

    Xp[mut] = np.random.normal(X[mut], sigma * (_xu * _xl))

    Xp = repair_random_init(Xp, X, xl, xu)

    return Xp


# ---------------------------------------------------------------------------------------------------------
# Class
# ---------------------------------------------------------------------------------------------------------


class GaussianMutation(Mutation):

    def __init__(self, sigma=0.1, **kwargs):
        super().__init__(**kwargs)
        self.sigma = Real(sigma, bounds=(0.01, 0.25), strict=(0.0, 1.0))

    def _do(self, problem, X, **kwargs):
        X = X.astype(float)

        sigma = get(self.sigma, size=len(X))
        prob_var = self.get_prob_var(problem, size=len(X))

        Xp = mut_gauss(X, problem.xl, problem.xu, sigma, prob_var)

        return Xp


class GM(GaussianMutation):
    pass



import abc

import numpy as np


class Operator:

    def __init__(self,
                 name=None,
                 vtype=None,
                 repair=None) -> None:

        super().__init__()

        if name is None:
            name = self.__class__.__name__

        self.name = name
        self.vtype = vtype
        self.repair = repair

    @abc.abstractmethod
    def do(self, problem, elem, *args, **kwargs):
        pass

    def __call__(self, problem, elem, *args, to_numpy=False, **kwargs):
        out = self.do(problem, elem, *args, **kwargs)

        if self.vtype is not None:
            for ind in out:
                ind.X = ind.X.astype(self.vtype)

        # allow to have a built-in repair (can be useful to customize standard crossover)
        if self.repair is not None:
            self.repair.do(problem, out)

        if to_numpy:
            out = np.array([ind.X for ind in out])

        return out



from pymoo.util.misc import swap

import numpy as np


def stochastic_ranking(f, phi, pr, I=None):
    _lambda = len(f)

    if I is None:
        I = np.arange(_lambda)

    for i in range(_lambda):

        at_least_one_swap = False

        for j in range(_lambda - 1):

            u = np.random.random()

            if u < pr or (phi[I[j]] == 0 and phi[I[j + 1]] == 0):
                if f[I[j]] > f[I[j + 1]]:
                    swap(I, j, j + 1)
                    at_least_one_swap = True
            else:
                if phi[I[j]] > phi[I[j + 1]]:
                    swap(I, j, j + 1)
                    at_least_one_swap = True

        if not at_least_one_swap:
            break

    return I



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True


import numpy as np
from libcpp cimport bool
from libcpp.vector cimport vector


cdef extern from "math.h":
    cpdef double floor(double x)

cdef extern from "limits.h":
    int INT_MAX




# ---------------------------------------------------------------------------------------------------------
# Interface
# ---------------------------------------------------------------------------------------------------------



def fast_non_dominated_sort(double[:,:] F, double epsilon = 0.0, int n_stop_if_ranked=INT_MAX):
    return c_fast_non_dominated_sort(F, epsilon, n_stop_if_ranked)

def best_order_sort(double[:,:] F):
    return c_best_order_sort(F)

def get_relation(F, a, b):
    return c_get_relation(F, a, b)

def fast_best_order_sort(double[:,:] F):
    return c_fast_best_order_sort(F)

def efficient_non_dominated_sort(double[:,:] F, strategy="sequential"):
    assert (strategy in ["sequential", 'binary']), "Invalid search strategy"
    return c_efficient_non_dominated_sort(F, strategy)





# ---------------------------------------------------------------------------------------------------------
# Fast Non-Dominated Sort
# ---------------------------------------------------------------------------------------------------------



cdef vector[vector[int]] c_fast_non_dominated_sort(double[:,:] F, double epsilon = 0.0, int n_stop_if_ranked=INT_MAX):

    cdef:
        int n_points, i, j, rel, n_ranked
        vector[int] current_front, next_front, n_dominated
        vector[vector[int]] fronts, is_dominating

    # calculate the dominance matrix
    n_points = F.shape[0]

    fronts = vector[vector[int]]()

    if n_points == 0:
        return fronts

    # final rank that will be returned
    n_ranked = 0

    # for each individual a list of all individuals that are dominated by this one
    is_dominating = vector[vector[int]](n_points)
    for _ in range(n_points):
        is_dominating.push_back(vector[int](n_points))

    n_dominated = vector[int]()
    for i in range(n_points):
        n_dominated.push_back(0)

    current_front = vector[int]()

    for i in range(n_points):

        for j in range(i + 1, n_points):

            rel = c_get_relation(F, i, j, epsilon)

            if rel == 1:
                is_dominating[i].push_back(j)
                n_dominated[j] += 1

            elif rel == -1:
                is_dominating[j].push_back(i)
                n_dominated[i] += 1

        if n_dominated[i] == 0:
            current_front.push_back(i)
            n_ranked += 1

    # append the first front to the current front
    fronts.push_back(current_front)

    # while not all solutions are assigned to a pareto front or we can stop early because of stop criterium
    while (n_ranked < n_points) and (n_ranked < n_stop_if_ranked):

        next_front = vector[int]()

        # for each individual in the current front
        for i in current_front:

            # all solutions that are dominated by this individuals
            for j in is_dominating[i]:

                n_dominated[j] -= 1
                if n_dominated[j] == 0:
                    next_front.push_back(j)
                    n_ranked += 1

        fronts.push_back(next_front)
        current_front = next_front

    return fronts


cdef vector[vector[int]] c_best_order_sort(double[:,:] F):

    cdef:
        int n_points, n_obj, n_fronts, n_ranked, i, j, s, e, l, z
        vector[int] rank
        int[:,:] Q
        bool is_dominated
        vector[vector[int]] fronts, empty
        vector[vector[vector[int]]] L

    n_points = F.shape[0]
    n_obj = F.shape[1]
    fronts = vector[vector[int]]()

    _Q = np.zeros((n_points, n_obj), dtype=np.intc)
    for j in range(n_obj):
        _Q[:, j] = np.lexsort(F[:, j:][:, ::-1].T, axis=0)
    Q = _Q[:,:]

    rank = vector[int]()
    for i in range(n_points):
        rank.push_back(-1)

    L = vector[vector[vector[int]]]()
    for j in range(n_obj):
        empty = vector[vector[int]]()
        L.push_back(empty)

    n_fronts = 0
    n_ranked = 0

    # the outer loop iterates through all solutions
    for i in range(n_points):

        # the inner loop through each objective values (sorted)
        for j in range(n_obj):

            # index of the current solution
            s = Q[i, j]

            # if solution was already ranked before - just append it to the corresponding front
            if rank[s] != -1:
                L[j][rank[s]].push_back(s)

            # otherwise we rank it for the first time
            else:

                # the rank of this solution is stored here
                s_rank = -1

                # for each front ranked for this objective
                for k in range(n_fronts):

                    is_dominated = False

                    # for each entry in that front
                    for e in L[j][k]:

                        is_dominated = c_get_relation(F, s, e) == -1

                        # if one solution dominates the current one - go to the next front
                        if is_dominated:
                            break

                    # if no solutions in the front dominates this one we found the rank
                    if not is_dominated:
                        s_rank = k
                        break

                # we need to add a new front for each objective
                if s_rank == -1:

                    s_rank = n_fronts
                    n_fronts += 1

                    fronts.push_back(vector[int]())
                    for l in range(n_obj):
                        L[l].push_back(vector[int]())

                L[j][s_rank].push_back(s)
                fronts[s_rank].push_back(s)
                rank[s] = s_rank
                n_ranked += 1

                if n_ranked == n_points:
                    break

    return fronts


cdef vector[vector[int]] c_fast_best_order_sort(double[:,:] F):

    cdef:
        int n_points, n_obj, n_fronts, n_ranked, i, j, s, e, l, z, s_next
        vector[int] rank, counter, check_if_equal
        int[:,:] Q
        bool is_dominated
        vector[vector[int]] fronts, empty ,C
        vector[vector[vector[int]]] L

    n_points = F.shape[0]
    n_obj = F.shape[1]
    fronts = vector[vector[int]]()


    _Q = np.zeros((n_points, n_obj), dtype=np.intc)
    _Q[:, 0] = np.lexsort(F[:, ::-1].T, axis=0)
    for j in range(1, n_obj):
        _Q[:, j] = np.lexsort(np.vstack([_Q[:, 0], F[:, j]]), axis=0)
    Q = _Q[:,:]


    counter = vector[int](n_points)
    C = vector[vector[int]]()
    for j in range(n_points):
        C.push_back(vector[int](n_obj))
    for i in range(n_points):
        for j in range(n_obj):
            s = Q[i, j]
            C[s][counter[s]] = j
            counter[s] += 1
    counter = vector[int](n_points)

    check_if_equal = vector[int]()
    for j in range(n_points):
        check_if_equal.push_back(-1)

    rank = vector[int]()
    for i in range(n_points):
        rank.push_back(-1)

    L = vector[vector[vector[int]]]()
    for j in range(n_obj):
        empty = vector[vector[int]]()
        L.push_back(empty)

    n_fronts = 0
    n_ranked = 0

    # the outer loop iterates through all solutions
    for i in range(n_points):

        # the inner loop through each objective values (sorted)
        for j in range(n_obj):

            # index of the current solution
            s = Q[i, j]
            s_next = Q[i + 1, j]

            # increase the counter for comparing this objective
            counter[s] += 1

            # if not the last solution
            if i < n_points - 1:
                if check_if_equal[s] == -1:
                    check_if_equal[s] = s_next
                elif check_if_equal[s] != s_next:
                    check_if_equal[s] = -2

            # if solution was already ranked before - just append it to the corresponding front
            if rank[s] != -1:
                L[j][rank[s]].push_back(s)

            # otherwise we rank it for the first time
            else:

                # the rank of this solution is stored here
                s_rank = -1

                # for each front ranked for this objective
                for k in range(n_fronts):

                    # just necessary if no fronts exists
                    is_dominated = False
                    is_equal = False

                    # for each entry in that front
                    for e in L[j][k]:

                        # get the domination relation - might return true even if equal
                        is_dominated = c_is_dominating_or_equal(F, e, s, C, counter[s])

                        if is_dominated and check_if_equal[e] == s:
                            is_equal = c_is_equal(F, s, e)

                        # if just one solution dominates the current one - go to the next front
                        if is_dominated or is_equal:
                            break

                    # if no solutions in the front dominates this one we found the rank
                    if not is_dominated or is_equal:
                        s_rank = k
                        break

                # we need to add a new front for each objective
                if s_rank == -1:

                    s_rank = n_fronts
                    n_fronts += 1

                    fronts.push_back(vector[int]())
                    for l in range(n_obj):
                        L[l].push_back(vector[int]())

                L[j][s_rank].push_back(s)
                fronts[s_rank].push_back(s)
                rank[s] = s_rank
                n_ranked += 1

                if n_ranked == n_points:
                    break

    return fronts




# ---------------------------------------------------------------------------------------------------------
# Efficient Non-dominated Sort
# ---------------------------------------------------------------------------------------------------------


cdef vector[vector[int]] c_efficient_non_dominated_sort(double[:,:] F, str strategy):
    cdef:
        int i, j, k, n, val
        vector[int] empty, e
        vector[vector[int]] fronts, ret

    # number of individuals
    n = len(F)

    # sort the input lexicographically
    indices = np.lexsort(F.T[::-1])
    F = np.asarray(F)[indices]

    # the fronts to be set for each iteration
    fronts = vector[vector[int]]()

    for i in range(n):

        if strategy == "sequential":
            k = sequential_search(F, i, fronts)
        else:
            k = binary_search(F, i, fronts)

        if k >= fronts.size():
            empty = vector[int]()
            fronts.push_back(empty)

        fronts[k].push_back(i)

    # convert to the return array
    ret = vector[vector[int]]()
    for i in range(fronts.size()):
        e = vector[int]()
        for j in range(fronts[i].size()):
            k = fronts[i][j]
            val = indices[k]
            e.push_back(val)
        ret.push_back(e)

    return ret



cdef int sequential_search(double[:,:] F, int i, vector[vector[int]] fronts):

    cdef:
        int k, j, n_fronts
        bool non_dominated

    n_fronts = fronts.size()
    if n_fronts == 0:
        return 0

    k = 0
    while True:

        non_dominated = True

        # solutions in the k-th front, examine in reverse order
        j = fronts[k].size() - 1

        while j >= 0:
            relation = c_get_relation(F, i, fronts[k][j])
            if relation == -1:
                non_dominated = False
                break
            j = j - 1

        if non_dominated:
            return k

        # move the individual to a new front
        else:
            k += 1
            if k >= n_fronts:
                return n_fronts


cdef int binary_search(double[:,:] F, int i, vector[vector[int]] fronts):

    cdef:
        int n_fronts, k, k_min, k_max, j
        bool non_dominated

    n_fronts = fronts.size()
    if n_fronts == 0:
        return 0

    k_min = 0  # the lower bound for checking
    k_max = n_fronts  # the upper bound for checking
    k = int(floor((k_max + k_min) / 2.0 + 0.5))  # the front now checked

    while True:

        non_dominated = True

        # solutions in the k-th front, examine in reverse order
        j = fronts[k-1].size() - 1

        while j >= 0:
            relation = c_get_relation(F, i, fronts[k-1][j])
            if relation == -1:
                non_dominated = False
                break
            j = j - 1

        # binary search
        if non_dominated:
            if k == k_min + 1:
                return k - 1
            else:
                k_max = k
                k = int(floor((k_max + k_min) / 2.0 + 0.5))

        else:
            k_min = k
            if k_max == k_min + 1 and k_max < n_fronts:
                return k_max - 1
            elif k_min == n_fronts:
                return n_fronts
            else:
                k = int(floor((k_max + k_min) / 2.0 + 0.5))






# ---------------------------------------------------------------------------------------------------------
# Util
# ---------------------------------------------------------------------------------------------------------


cdef int c_get_relation(double[:,:] F, int a, int b, double epsilon = 0.0):

    cdef int size = F.shape[1]
    cdef int val
    cdef int i

    val = 0

    for i in range(size):

        if F[a,i] + epsilon < F[b,i]:
            # indifferent because once better and once worse
            if val == -1:
                return 0
            val = 1

        elif F[a,i] > F[b,i] + epsilon:

            # indifferent because once better and once worse
            if val == 1:
                return 0
            val = -1

    return val



cdef bool c_is_dominating_or_equal(double[:,:] F, int a, int b, vector[vector[int]]& C, int k):
    cdef int i, j
    for i in range(k, C[0].size()):
        j = C[b][i]
        if F[b, j] < F[a, j]:
            return False
    return True

cdef bool c_is_equal(double[:,:] F, int a, int b):
    cdef int i
    cdef int n_obj = F.shape[1]
    for i in range(n_obj):
        if F[a, i] != F[b, i]:
            return False
    return True



import numpy as np

from pymoo.core.duplicate import DefaultDuplicateElimination
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.survival import Survival


def is_better(_new, _old, eps=0.0):
    both_infeasible = not _old.feasible[0] and not _new.feasible[0]
    both_feasible = _old.feasible[0] and _new.feasible[0]

    if both_infeasible and _old.CV[0] - _new.CV[0] > eps:
        return True
    elif not _old.feasible and _new.feasible:
        return True
    elif both_feasible and _old.F[0] - _new.F[0] > eps:
        return True

    return False


class ReplacementSurvival(Survival):

    def do(self, problem, pop, off, return_indices=False, inplace=False, **kwargs):

        # this makes it usable as a traditional survival
        if isinstance(off, int):
            k = off
            off = pop[k:]
            pop = pop[:k]

        # if the offsprings are simply empty don't do anything
        if len(off) == 0:
            return pop

        assert len(pop) == len(off), "For the replacement pop and off must have the same number of individuals."

        pop = Population.create(pop) if isinstance(pop, Individual) else pop
        off = Population.create(off) if isinstance(off, Individual) else off

        I = self._do(problem, pop, off, **kwargs)

        if return_indices:
            return I
        else:
            if not inplace:
                pop = pop.copy()
            pop[I] = off[I]
            return pop

    def _do(self, problem, pop, off, **kwargs):
        pass


class ImprovementReplacement(ReplacementSurvival):

    def _do(self, problem, pop, off, **kwargs):

        ret = np.full((len(pop), 1), False)

        pop_F, pop_CV, pop_feas = pop.get("F", "CV", "feasible")
        off_F, off_CV, off_feas = off.get("F", "CV", "feasible")

        if problem.has_constraints() > 0:

            # 1) Both infeasible and constraints have been improved
            ret[(~pop_feas & ~off_feas) & (off_CV < pop_CV)] = True

            # 2) A solution became feasible
            ret[~pop_feas & off_feas] = True

            # 3) Both feasible but objective space value has improved
            ret[(pop_feas & off_feas) & (off_F < pop_F)] = True

        else:
            ret[off_F < pop_F] = True

        # never allow duplicates to become part of the population when replacement is used
        _, _, is_duplicate = DefaultDuplicateElimination(epsilon=0.0).do(off, pop, return_indices=True)
        ret[is_duplicate] = False

        return ret[:, 0]


def parameter_less(f, cv):
    v = np.copy(f)
    infeas = cv > 0
    v[infeas] = f.max() + cv[infeas]
    return v


def hierarchical_sort(f, cv=None):
    if cv is not None:
        f = parameter_less(f, cv)
    return np.argsort(f)



from math import floor

import numpy as np

from pymoo.util.dominator import Dominator


def efficient_non_dominated_sort(F, strategy="sequential"):
    """
    Efficient Non-dominated Sorting (ENS)

    Parameters
    ----------
    F: numpy.ndarray
        objective values for each individual.
    strategy: str
        search strategy, can be "sequential" or "binary".

    Returns
    -------
        fronts: list
            Indices of the individuals in each front.

    References
    ----------
    X. Zhang, Y. Tian, R. Cheng, and Y. Jin,
    An efficient approach to nondominated sorting for evolutionary multiobjective optimization,
    IEEE Transactions on Evolutionary Computation, 2015, 19(2): 201-213.
    """

    assert (strategy in ["sequential", 'binary']), "Invalid search strategy"

    # the shape of the input
    N, M = F.shape

    # do a lexicographic ordering
    I = np.lexsort(F.T[::-1])
    F = F[I]

    # front ranks for each individual
    fronts = []

    for i in range(N):

        if strategy == 'sequential':
            k = sequential_search(F, i, fronts)
        else:
            k = binary_search(F, i, fronts)

        # create empty fronts if necessary
        if k >= len(fronts):
            fronts.append([])

        # append the current individual to a front
        fronts[k].append(i)

    # now map the fronts back to the originally sorting
    ret = []
    for front in fronts:
        ret.append(I[front])

    return ret


def sequential_search(F, i, fronts) -> int:
    """
    Find the front rank for the i-th individual through sequential search.

    Parameters
    ----------
    F: np.ndarray
        the objective values
    i: int
        the index of the individual
    fronts: list
        individuals in each front
    """

    num_found_fronts = len(fronts)
    k = 0  # the front now checked
    current = F[i]
    while True:
        if num_found_fronts == 0:
            return 0
        # solutions in the k-th front, examine in reverse order
        fk_indices = fronts[k]
        solutions = F[fk_indices[::-1]]
        non_dominated = True
        for f in solutions:
            relation = Dominator.get_relation(current, f)
            if relation == -1:
                non_dominated = False
                break
        if non_dominated:
            return k
        else:
            k += 1
            if k >= num_found_fronts:
                # move the individual to a new front
                return num_found_fronts


def binary_search(F, i, fronts):
    """
    Find the front rank for the i-th individual through binary search.

    Parameters
    ----------
    F: np.ndarray
        the objective values
    i: int
        the index of the individual
    fronts: list
        individuals in each front
    """

    num_found_fronts = len(fronts)
    if num_found_fronts == 0:
        return 0

    k_min = 0  # the lower bound for checking
    k_max = num_found_fronts  # the upper bound for checking
    k = floor((k_max + k_min) / 2 + 0.5)  # the front now checked
    current = F[i]
    while True:

        # solutions in the k-th front, examine in reverse order
        fk_indices = fronts[k - 1]
        solutions = F[fk_indices[::-1]]
        non_dominated = True

        for f in solutions:
            relation = Dominator.get_relation(current, f)
            if relation == -1:
                non_dominated = False
                break

        # binary search
        if non_dominated:
            if k == k_min + 1:
                return k - 1
            else:
                k_max = k
                k = floor((k_max + k_min) / 2 + 0.5)
        else:
            k_min = k
            if k_max == k_min + 1 and k_max < num_found_fronts:
                return k_max - 1
            elif k_min == num_found_fronts:
                return num_found_fronts
            else:
                k = floor((k_max + k_min) / 2 + 0.5)




def get_termination(name, *args, **kwargs):
    from pymoo.termination.default import DefaultMultiObjectiveTermination, DefaultSingleObjectiveTermination
    from pymoo.termination.max_eval import MaximumFunctionCallTermination
    from pymoo.termination.max_gen import MaximumGenerationTermination
    from pymoo.termination.max_time import TimeBasedTermination
    from pymoo.termination.fmin import MinimumFunctionValueTermination

    TERMINATION = {
        "n_eval": MaximumFunctionCallTermination,
        "n_evals": MaximumFunctionCallTermination,
        "n_gen": MaximumGenerationTermination,
        "n_iter": MaximumGenerationTermination,
        "fmin": MinimumFunctionValueTermination,
        "time": TimeBasedTermination,
        "soo": DefaultSingleObjectiveTermination,
        "moo": DefaultMultiObjectiveTermination,
    }

    if name not in TERMINATION:
        raise Exception("Termination not found.")

    return TERMINATION[name](*args, **kwargs)



from os.path import dirname, realpath

from pymoo.version import __version__


class Config:
    """
    The configuration of this package in general providing the place
    for declaring global variables.
    """

    # the root directory where the package is located at
    root = dirname(realpath(__file__))

    warnings = {
        "not_compiled": True
    }

    # whether a warning should be printed if compiled modules are not available
    show_compile_hint = True

    # whether when import a file the doc should be parsed - only activate when creating doc files
    parse_custom_docs = False

    # a method defining the endpoint to load data remotely - default from github repo
    @classmethod
    def data(cls):
        return f"https://raw.githubusercontent.com/anyoptimization/pymoo-data/main/"


# returns the directory to be used for imports
def get_pymoo():
    return dirname(Config.root)



import numpy as np

from pymoo.algorithms.base.local import LocalSearch
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.replacement import is_better
from pymoo.docs import parse_doc_string
from pymoo.operators.repair.to_bound import set_to_bounds_if_outside_by_problem
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.optimum import filter_optimum


# =========================================================================================================
# Implementation
# =========================================================================================================


class PatternSearch(LocalSearch):
    def __init__(self,
                 init_delta=0.25,
                 init_rho=0.5,
                 step_size=1.0,
                 output=SingleObjectiveOutput(),
                 **kwargs):
        """
        An implementation of well-known Hooke and Jeeves Pattern Search.

        Parameters
        ----------

        x0 : numpy.array
            The initial value where the local search should be initiated. If not provided `n_sample_points` are
            created using latin hypercube sampling and the best solution found is set to `x0`.

        n_sample_points : int
            Number of sample points to be used to determine the initial search point. (Only used of `x0` is not provided)

        delta : float
            The `delta` values which is used for the exploration move. If lower and upper bounds are provided the
            value is in relation to the overall search space. For instance, a value of 0.25 means that initially the
            pattern is created in 25% distance of the initial search point.

        rho : float
            If the move was unsuccessful then the `delta` value is reduced by multiplying it with the value provided.
            For instance, `explr_rho` implies that with a value of `delta/2` is continued.

        step_size : float
            After the exploration move the new center is determined by following a promising direction.
            This value defines how large to step on this direction will be.

        """

        super().__init__(output=output, **kwargs)
        self.init_rho = init_rho
        self.init_delta = init_delta
        self.step_size = step_size

        self.n_not_improved = 0

        self._rho = init_rho
        self._delta = None
        self._center = None
        self._current = None
        self._trial = None
        self._direction = None
        self._sign = None

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills=infills, **kwargs)
        self._center, self._explr = self.x0, self.x0
        self._sign = np.ones(self.problem.n_var)

        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            self._delta = self.init_delta * (xu - xl)
        else:
            self._delta = np.abs(self.x0.X) / 2.0
            self._delta[self._delta <= 1.0] = 1.0

    def _next(self):

        # whether the last iteration has resulted in a new optimum or not
        has_improved = is_better(self._explr, self._center, eps=0.0)

        # that means that the exploration did not find any new point and was thus unsuccessful
        if not has_improved:

            # increase the counter (by default this will be initialized to 0 and directly increased to 1)
            self.n_not_improved += 1

            # keep track of the rho values in the normalized space
            self._rho = self.init_rho ** self.n_not_improved

            # explore around the current center - try finding a suitable direction
            self._explr = yield from exploration_move(self.problem, self._center, self._sign, self._delta, self._rho)

        # if we have found a direction in the last iteration to be worth following
        else:

            # get the direction which was successful in the last move
            self._direction = (self._explr.X - self._center.X)

            # declare the exploration point the new center (it has led to an improvement in the last iteration!)
            self._center = self._explr

            # use the pattern move to get a new trial vector along that given direction
            self._trial = yield pattern_move(self.problem, self._center, self._direction, self.step_size)

            # get the delta sign adjusted for the exploration
            self._sign = calc_sign(self._direction)

            # explore around the current center to try finding a suitable direction
            self._explr = yield from exploration_move(self.problem, self._trial, self._sign, self._delta, self._rho)

        self.pop = Population.create(self._center, self._explr)

    def _set_optimum(self):
        pop = self.pop if self.opt is None else Population.merge(self.opt, self.pop)
        self.opt = filter_optimum(pop, least_infeasible=True)


def exploration_move(problem, center, sign, delta, rho, randomize=True):
    n_var = problem.n_var

    # the order for the variable iteration
    if randomize:
        K = np.random.permutation(n_var)
    else:
        K = np.arange(n_var)

    # iterate over each variable
    for k in K:

        # the value to be tried first is given by the amount times the sign
        _delta = sign[k] * rho * delta

        # make a step of delta on the k-th variable
        _explr = yield step_along_axis(problem, center.X, _delta, k)

        if is_better(_explr, center, eps=0.0):
            center = _explr

        # if not successful try the other direction
        else:

            # now try the negative value of delta and see if we can improve
            _explr = yield step_along_axis(problem, center.X, -1 * _delta, k)

            if is_better(_explr, center, eps=0.0):
                center = _explr

    return center


def pattern_move(problem, current, direction, step_size):
    # calculate the new X and repair out of bounds if necessary
    X = current.X + step_size * direction
    set_to_bounds_if_outside_by_problem(problem, X)

    # create the new center individual
    return Individual(X=X)


def calc_sign(direction):
    sign = np.sign(direction)
    sign[sign == 0] = -1
    return sign


def step_along_axis(problem, x, delta, axis):
    # copy and add delta to the new point
    X = np.copy(x)

    # now add to the current solution
    X[axis] = X[axis] + delta[axis]

    # repair if out of bounds if necessary
    X = set_to_bounds_if_outside_by_problem(problem, X)

    return Individual(X=X)


parse_doc_string(PatternSearch.__init__)



import numpy as np

from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot
from pymoo.util.misc import set_if_none
from pymoo.visualization.util import get_circle_points, plot_axes_lines, \
    plot_axis_labels, plot_circle, plot_polygon, parse_bounds, normalize, equal_axis, no_ticks


class Petal(Plot):

    def __init__(self,
                 bounds=None,
                 **kwargs):
        """

        Petal Diagram


        Parameters
        ----------------
        bounds : tuple
            The boundaries for each objective. Necessary to be provided for this plot!
        axis_style : {axis_style}
        reverse : bool
            Default false. Otherwise, larger area means smaller value.

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}

        """

        super().__init__(bounds=bounds, **kwargs)

        if bounds is None:
            raise Exception("Boundaries must be provided for Petal Width. Otherwise, no trade-offs can be calculated.")

        set_if_none(self.axis_style, "color", "black")
        set_if_none(self.axis_style, "linewidth", 2)
        set_if_none(self.axis_style, "alpha", 0.5)

    def _plot(self, ax, F):

        # equal axis length and no ticks
        equal_axis(ax)
        no_ticks(ax)

        V = get_circle_points(len(F))

        # sections to plot
        sections = np.linspace(0, 2 * np.pi, self.n_dim + 1)

        t = [(sections[i] + sections[i + 1]) / 2 for i in range(len(sections) - 1)]
        endpoints = np.column_stack([np.cos(t), np.sin(t)])
        plot_axis_labels(ax, endpoints, self.get_labels(), **self.axis_label_style)

        center = np.zeros(2)

        for i in range(len(sections) - 1):
            t = np.linspace(sections[i], sections[i + 1], 100)
            v = np.column_stack([np.cos(t), np.sin(t)])

            P = np.row_stack([center, F[i] * v])
            plot_polygon(ax, P, color=self.colors[i])

        # draw the outer circle
        plot_circle(ax, **self.axis_style)
        plot_axes_lines(ax, V, **self.axis_style)

    def _do(self):

        n_rows = len(self.to_plot)
        n_cols = max([len(e[0]) for e in self.to_plot])
        self.init_figure(n_rows=n_rows, n_cols=n_cols, force_axes_as_matrix=True)

        # normalize the input
        bounds = parse_bounds(self.bounds, self.n_dim)
        to_plot_norm = normalize(self.to_plot, bounds, reverse=self.reverse)

        for k, (F, kwargs) in enumerate(to_plot_norm):
            for j, _F in enumerate(F):
                self._plot(self.ax[k, j], _F)


parse_doc_string(Petal.__init__)



#    Copyright (C) 2010 Simon Wessing
#    TU Dortmund University
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


__author__ = "Simon Wessing"


class HyperVolume:
    """
    Hypervolume computation based on cross 3 of the algorithm in the paper:
    C. M. Fonseca, L. Paquete, and M. Lopez-Ibanez. An improved dimension-sweep
    algorithm for the hypervolume indicator. In IEEE Congress on Evolutionary
    Computation, pages 1157-1163, Vancouver, Canada, July 2006.

    Minimization is implicitly assumed here!

    """

    def __init__(self, referencePoint):
        """Constructor."""
        self.referencePoint = referencePoint
        self.list = []

    def compute(self, front):
        """Returns the hypervolume that is dominated by a non-dominated front.

        Before the HV computation, front and reference point are translated, so
        that the reference point is [0, ..., 0].

        """

        def weaklyDominates(point, other):
            for i in range(len(point)):
                if point[i] > other[i]:
                    return False
            return True

        relevantPoints = []
        referencePoint = self.referencePoint
        dimensions = len(referencePoint)
        for point in front:
            # only consider points that dominate the reference point
            if weaklyDominates(point, referencePoint):
                relevantPoints.append(point)
        if any(referencePoint):
            # shift points so that referencePoint == [0, ..., 0]
            # this way the reference point doesn't have to be explicitly used
            # in the HV computation
            for j in range(len(relevantPoints)):
                relevantPoints[j] = [relevantPoints[j][i] - referencePoint[i] for i in range(dimensions)]
        self.preProcess(relevantPoints)
        bounds = [-1.0e308] * dimensions
        hyperVolume = self.hvRecursive(dimensions - 1, len(relevantPoints), bounds)
        return hyperVolume

    def hvRecursive(self, dimIndex, length, bounds):
        """Recursive call to hypervolume calculation.

        In contrast to the paper, the code assumes that the reference point
        is [0, ..., 0]. This allows the avoidance of a few operations.

        """
        hvol = 0.0
        sentinel = self.list.sentinel
        if length == 0:
            return hvol
        elif dimIndex == 0:
            # special case: only one dimension
            # why using hypervolume at all?
            return -sentinel.next[0].cargo[0]
        elif dimIndex == 1:
            # special case: two dimensions, end recursion
            q = sentinel.next[1]
            h = q.cargo[0]
            p = q.next[1]
            while p is not sentinel:
                pCargo = p.cargo
                hvol += h * (q.cargo[1] - pCargo[1])
                if pCargo[0] < h:
                    h = pCargo[0]
                q = p
                p = q.next[1]
            hvol += h * q.cargo[1]
            return hvol
        else:
            remove = self.list.remove
            reinsert = self.list.reinsert
            hvRecursive = self.hvRecursive
            p = sentinel
            q = p.prev[dimIndex]
            while q.cargo is not None:
                if q.ignore < dimIndex:
                    q.ignore = 0
                q = q.prev[dimIndex]
            q = p.prev[dimIndex]
            while length > 1 and (
                    q.cargo[dimIndex] > bounds[dimIndex] or q.prev[dimIndex].cargo[dimIndex] >= bounds[dimIndex]):
                p = q
                remove(p, dimIndex, bounds)
                q = p.prev[dimIndex]
                length -= 1
            qArea = q.area
            qCargo = q.cargo
            qPrevDimIndex = q.prev[dimIndex]
            if length > 1:
                hvol = qPrevDimIndex.volume[dimIndex] + qPrevDimIndex.area[dimIndex] * (
                        qCargo[dimIndex] - qPrevDimIndex.cargo[dimIndex])
            else:
                qArea[0] = 1
                qArea[1:dimIndex + 1] = [qArea[i] * -qCargo[i] for i in range(dimIndex)]
            q.volume[dimIndex] = hvol
            if q.ignore >= dimIndex:
                qArea[dimIndex] = qPrevDimIndex.area[dimIndex]
            else:
                qArea[dimIndex] = hvRecursive(dimIndex - 1, length, bounds)
                if qArea[dimIndex] <= qPrevDimIndex.area[dimIndex]:
                    q.ignore = dimIndex
            while p is not sentinel:
                pCargoDimIndex = p.cargo[dimIndex]
                hvol += q.area[dimIndex] * (pCargoDimIndex - q.cargo[dimIndex])
                bounds[dimIndex] = pCargoDimIndex
                reinsert(p, dimIndex, bounds)
                length += 1
                q = p
                p = p.next[dimIndex]
                q.volume[dimIndex] = hvol
                if q.ignore >= dimIndex:
                    q.area[dimIndex] = q.prev[dimIndex].area[dimIndex]
                else:
                    q.area[dimIndex] = hvRecursive(dimIndex - 1, length, bounds)
                    if q.area[dimIndex] <= q.prev[dimIndex].area[dimIndex]:
                        q.ignore = dimIndex
            hvol -= q.area[dimIndex] * q.cargo[dimIndex]
            return hvol

    def preProcess(self, front):
        """Sets up the list data structure needed for calculation."""
        dimensions = len(self.referencePoint)
        nodeList = MultiList(dimensions)
        nodes = [MultiList.Node(dimensions, point) for point in front]
        for i in range(dimensions):
            self.sortByDimension(nodes, i)
            nodeList.extend(nodes, i)
        self.list = nodeList

    def sortByDimension(self, nodes, i):
        """Sorts the list of nodes by the i-th value of the contained points."""
        # build a list of tuples of (point[i], node)
        decorated = [(node.cargo[i], index, node) for index, node in enumerate(nodes)]
        # sort by this value
        decorated.sort()
        # write back to original list
        nodes[:] = [node for (_, _, node) in decorated]


class MultiList:
    """A special data structure needed by FonsecaHyperVolume.

    It consists of several doubly linked lists that share common nodes. So,
    every node has multiple predecessors and successors, one in every list.

    """

    class Node:

        def __init__(self, numberLists, cargo=None):
            self.cargo = cargo
            self.next = [None] * numberLists
            self.prev = [None] * numberLists
            self.ignore = 0
            self.area = [0.0] * numberLists
            self.volume = [0.0] * numberLists

        def __str__(self):
            return str(self.cargo)

    def __init__(self, numberLists):
        """Constructor.

        Builds 'numberLists' doubly linked lists.

        """
        self.numberLists = numberLists
        self.sentinel = MultiList.Node(numberLists)
        self.sentinel.next = [self.sentinel] * numberLists
        self.sentinel.prev = [self.sentinel] * numberLists

    def __str__(self):
        strings = []
        for i in range(self.numberLists):
            currentList = []
            node = self.sentinel.next[i]
            while node != self.sentinel:
                currentList.append(str(node))
                node = node.next[i]
            strings.append(str(currentList))
        stringRepr = ""
        for string in strings:
            stringRepr += string + "\n"
        return stringRepr

    def __len__(self):
        """Returns the number of lists that are included in this MultiList."""
        return self.numberLists

    def getLength(self, i):
        """Returns the length of the i-th list."""
        length = 0
        sentinel = self.sentinel
        node = sentinel.next[i]
        while node != sentinel:
            length += 1
            node = node.next[i]
        return length

    def append(self, node, index):
        """Appends a node to the end of the list at the given index."""
        lastButOne = self.sentinel.prev[index]
        node.next[index] = self.sentinel
        node.prev[index] = lastButOne
        # set the last element as the new one
        self.sentinel.prev[index] = node
        lastButOne.next[index] = node

    def extend(self, nodes, index):
        """Extends the list at the given index with the nodes."""
        sentinel = self.sentinel
        for node in nodes:
            lastButOne = sentinel.prev[index]
            node.next[index] = sentinel
            node.prev[index] = lastButOne
            # set the last element as the new one
            sentinel.prev[index] = node
            lastButOne.next[index] = node

    def remove(self, node, index, bounds):
        """Removes and returns 'node' from all lists in [0, 'index'[."""
        for i in range(index):
            predecessor = node.prev[i]
            successor = node.next[i]
            predecessor.next[i] = successor
            successor.prev[i] = predecessor
            if bounds[i] > node.cargo[i]:
                bounds[i] = node.cargo[i]
        return node

    def reinsert(self, node, index, bounds):
        """
        Inserts 'node' at the position it had in all lists in [0, 'index'[
        before it was removed. This method assumes that the next and previous
        nodes of the node that is reinserted are in the list.

        """
        for i in range(index):
            node.prev[i].next[i] = node
            node.next[i].prev[i] = node
            if bounds[i] > node.cargo[i]:
                bounds[i] = node.cargo[i]


if __name__ == "__main__":
    # Example:
    referencePoint = [2, 2, 2]
    hv = HyperVolume(referencePoint)
    front = [[1, 0, 1], [0, 1, 0]]
    volume = hv.compute(front)



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True

import numpy as np

cdef extern from "vendor/hypervolume.cpp":
    double overmars_yap(double * points, double * referencePoint, unsigned noObjectives, unsigned noPoints);


def hv(ref_point, F):
    F = F[np.argsort(F[:, -1])]
    n, m = F.shape
    return c_hv(F, ref_point, m, n)


def c_hv(double[:,:] F, double[:] ref_point, m, n):
    return overmars_yap(&F[0, 0], &ref_point[0], m, n)
    # return fpli_hv(&F[0,0], m, n, &ref_point[0])



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.variable import get, Real
from pymoo.util.misc import crossover_mask, row_at_least_once_true


def mut_exp(n_matings, n_var, prob, at_least_once=True):
    assert len(prob) == n_matings

    # the mask do to the crossover
    M = np.full((n_matings, n_var), False)

    # start point of crossover
    s = np.random.randint(0, n_var, size=n_matings)

    # create for each individual the crossover range
    for i in range(n_matings):

        # the actual index where we start
        start = s[i]
        for j in range(n_var):

            # the current position where we are pointing to
            current = (start + j) % n_var

            # replace only if random value keeps being smaller than CR
            if np.random.random() <= prob[i]:
                M[i, current] = True
            else:
                break

    if at_least_once:
        M = row_at_least_once_true(M)

    return M


class ExponentialCrossover(Crossover):

    def __init__(self, prob_exp=0.75, **kwargs):
        super().__init__(2, 2, **kwargs)
        self.prob_exp = Real(prob_exp, bounds=(0.5, 0.9), strict=(0.0, 1.0))

    def _do(self, _, X, **kwargs):
        _, n_matings, n_var = X.shape
        prob_exp = get(self.prob_exp, size=n_matings)

        M = mut_exp(n_matings, n_var, prob_exp, at_least_once=True)
        _X = crossover_mask(X, M)
        return _X



import numpy as np

from pymoo.core.individual import calc_cv


class KKTPM:

    def calc(self, X, problem, ideal=None, utopian_eps=1e-4, rho=1e-3):
        """

        Returns the Karush-Kuhn-Tucker Approximate Measure.

        Parameters
        ----------
        X : np.array

        problem : pymoo.core.problem
        ideal : np.array
            The ideal point if not in the problem defined or intentionally overwritten.
        utopian_eps : float
            The epsilon used for decrease the ideal point to get the utopian point.
        rho : float
            Since augmented achievement scalarization function is used the F for all other weights
            - here rho - needs to be defined.

        Returns
        -------

        """

        # the final result to be returned
        kktpm = np.full((X.shape[0], 1), np.inf)
        fval = np.full((X.shape[0], 1), np.inf)

        # set the ideal point for normalization
        z = ideal

        # if not provided take the one defined in the problem
        if z is None:
            z = problem.ideal_point()
        z -= utopian_eps

        # for convenience get the counts directly
        n_solutions, n_var, n_obj, n_ieq_constr = X.shape[0], problem.n_var, problem.n_obj, problem.n_ieq_constr

        F, G, dF, dG = problem.evaluate(X, return_values_of=["F", "G", "dF", "dG"])
        CV = calc_cv(G=G)

        # loop through each solution to be considered
        for i in range(n_solutions):

            # get the corresponding values for this solution
            x, f, cv, df = X[i, :], F[i, :], CV[i], dF[i, :].swapaxes(1, 0)
            if n_ieq_constr > 0:
                g, dg = G[i, :], dG[i].T

            # if the solution that is provided is infeasible
            if cv > 0:
                _kktpm = 1 + cv
                _fval = None

            else:

                w = np.sqrt(np.sum(np.power(f - z, 2))) / (f - z)
                a_m = (df * w + (rho * np.sum(df * w, axis=1))[:, None]).T

                A = np.ones((problem.n_obj, problem.n_obj)) + a_m @ a_m.T
                b = np.ones(problem.n_obj)

                if n_ieq_constr > 0:
                    # a_j is just the transpose of the differential of constraints
                    a_j = dg.T

                    # part of the matrix for additional constraints
                    gsq = np.zeros((n_ieq_constr, n_ieq_constr))
                    np.fill_diagonal(gsq, g * g)

                    # now add the constraints to the optimization problem
                    A = np.vstack([np.hstack([A, a_m @ a_j.T]), np.hstack([a_j @ a_m.T, a_j @ a_j.T + gsq])])
                    b = np.hstack([b, np.zeros(n_ieq_constr)])

                method = "qr"
                u = solve(A, b, method=method)

                # until all the lagrange multiplier are positive
                while np.any(u < 0):

                    # go through one by one
                    for j in range(len(u)):

                        # if a lagrange multiplier is negative - we need to fix it
                        if u[j] < 0:
                            # modify the optimization problem
                            A[j, :], A[:, j], A[j, j] = 0, 0, 1
                            b[j] = 0

                            # resolve the problem and redefine u. for sure all preview u[j] are positive now
                            u = solve(A, b, method=method)

                # split up the lagrange multiplier for objective and not
                u_m, u_j = u[:n_obj], u[n_obj:]

                if n_ieq_constr > 0:
                    _kktpm = (1 - np.sum(u_m)) ** 2 + np.sum((np.vstack([a_m, a_j]).T @ u) ** 2)
                    _fval = _kktpm + np.sum((u_j * g.T) ** 2)
                else:
                    _kktpm = (1 - np.sum(u_m)) ** 2 + np.sum((a_m.T @ u) ** 2)
                    _fval = _kktpm

                ujgj = -g @ u_j
                if np.sum(u_m) + ujgj * (1 + ujgj) > 1:
                    adjusted_kktpm = - (u_j @ g.T)
                    projected_kktpm = (_kktpm * g @ g.T - g @ u_j) / (1 + g @ g.T)
                    _kktpm = (_kktpm + adjusted_kktpm + projected_kktpm) / 3

            # assign to the values to be returned
            kktpm[i] = _kktpm
            fval[i] = _fval

        return kktpm[:, 0]


def solve(A, b, method="elim"):
    if method == "elim":
        return np.linalg.solve(A, b)

    elif method == "qr":
        Q, R = np.linalg.qr(A)
        y = np.dot(Q.T, b)
        return np.linalg.solve(R, y)

    elif method == "svd":
        U, s, V = np.linalg.svd(A)  # SVD decomposition of A
        A_inv = np.dot(np.dot(V.T, np.linalg.inv(np.diag(s))), U.T)
        return A_inv @ b


if __name__ == '__main__':
    from pymoo.problems import get_problem
    from pymoo.gradient.automatic import AutomaticDifferentiation

    from pymoo.constraints.from_bounds import ConstraintsFromBounds
    problem = ConstraintsFromBounds(AutomaticDifferentiation(get_problem("zdt2", n_var=30)))

    # X = (0.5 * np.ones(10))[None, :]
    X = np.array(
        [0.394876, 0.963263, 0.173956, 0.126330, 0.135079, 0.505662, 0.021525, 0.947970, 0.827115, 0.015019, 0.176196,
         0.332064, 0.130997, 0.809491, 0.344737, 0.940107, 0.582014, 0.878832, 0.844734, 0.905392, 0.459880, 0.546347,
         0.798604, 0.285719, 0.490254, 0.599110, 0.015533, 0.593481, 0.433676, 0.807361])

    print(KKTPM().calc(X[None, :], problem))



import numpy as np

from pymoo.problems.dyn import DynamicTestProblem


class DynamicCEC2015(DynamicTestProblem):

    def __init__(self, n_var=10, nt=10, taut=20, n_obj=2, xl=0.0, xu=1.0, vtype=float, **kwargs):
        super().__init__(nt, taut, n_var=n_var, n_obj=n_obj, xl=xl, xu=xu, vtype=vtype, **kwargs)


class FDA2DEB(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import fda2_deb as f
        out["F"] = np.array([f(x, t) for x in X])


class FDA4(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, n_obj=3, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import FDA4 as f
        out["F"] = np.array([f(x, t) for x in X])


class FDA5(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, n_obj=3, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import FDA5 as f
        out["F"] = np.array([f(x, t) for x in X])


class DIMP2(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import DIMP2 as f
        out["F"] = np.array([f(x, t) for x in X])


class dMOP2(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import dMOP2 as f
        out["F"] = np.array([f(x, t) for x in X])


class dMOP3(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import dMOP3 as f
        out["F"] = np.array([f(x, t) for x in X])


class HE2(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import HE2 as f
        out["F"] = np.array([f(x, t) for x in X])


class HE7(DynamicCEC2015):

    def __init__(self, n_var=10, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import HE7 as f
        out["F"] = np.array([f(x, t) for x in X])


class HE9(DynamicCEC2015):

    def __init__(self, n_var=10, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import HE9 as f
        out["F"] = np.array([f(x, t) for x in X])



import numpy as np
from scipy.spatial.ckdtree import cKDTree

from pymoo.core.indicator import Indicator


class DecisionMaking(Indicator):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.default_if_empty = None

    def _do(self, F, *args, **kwargs):
        pass


class NeighborFinder:

    def __init__(self, N,
                 epsilon=0.125,
                 n_neighbors=None,
                 n_min_neigbors=None,
                 consider_2d=True):

        super().__init__()
        self.N = N
        self.consider_2d = consider_2d

        _, n_dim = N.shape

        # at least find min(dimensionality times two neighbors, number PO solutions - 1) - if enabled
        if n_min_neigbors == "auto":
            self.n_min_neigbors = min(2 * n_dim, _ - 1)

        # disable the minimum neighbor variable
        else:
            self.n_min_neigbors = np.inf

        # either choose epsilon
        self.epsilon = epsilon

        # if none choose the number of neighbors
        self.n_neighbors = n_neighbors

        if self.N.shape[1] == 1:
            raise Exception("At least 2 objectives must be provided.")

        elif self.consider_2d and self.N.shape[1] == 2:
            self.min, self.max = N.min(), N.max()
            self.rank = np.argsort(N[:, 0])
            self.pos_in_rank = np.argsort(self.rank)

        else:
            self.tree = cKDTree(N)

    def find(self, i):

        if self.consider_2d and self.N.shape[1] == 2:
            neighbours = []

            pos = self.pos_in_rank[i]
            if pos > 0:
                neighbours.append(self.rank[pos - 1])
            if pos < len(self.N) - 1:
                neighbours.append(self.rank[pos + 1])

        else:

            # for each neighbour in a specific radius of that solution
            if self.epsilon is not None:
                neighbours = self.tree.query_ball_point([self.N[i]], self.epsilon).tolist()[0]
            elif self.n_neighbors is not None:
                neighbours = self.tree.query([self.N[i]], k=self.n_neighbors + 1)[1].tolist()[0]
            else:
                raise Exception("Either define epsilon or number of neighbors.")

            # in case n_min_neigbors is enabled
            if len(neighbours) < self.n_min_neigbors:
                neighbours = self.tree.query([self.N[i]], k=self.n_min_neigbors + 1)[1].tolist()[0]

        return neighbours


def find_outliers_upper_tail(mu):

    # remove values that are nan
    I = np.where(np.logical_and(np.logical_not(np.isnan(mu)), np.logical_not(np.isinf(mu))))[0]
    mu = mu[I]

    # calculate mean and sigma
    mean, sigma = mu.mean(), mu.std()

    # calculate the deviation in terms of sigma
    deviation = (mu - mean) / sigma

    # 2 * sigma is considered as an outlier
    S = I[np.where(deviation >= 2)[0]]

    if len(S) == 0 and deviation.max() > 1:
        S = I[[np.argmax(mu)]]

    return S if len(S) > 0 else None



from functools import partial

from jax.config import config

config.update("jax_enable_x64", True)

import pymoo.gradient.toolbox as anp
import numpy as np
from jax import vjp
from jax import vmap
from jax._src.api import _jacrev_unravel, _std_basis
from jax.tree_util import (tree_map)


def jax_elementwise_value_and_grad(f, x):
    out, pullback = vjp(f, x)
    u = _std_basis(out)
    jac, = vmap(pullback, in_axes=0)(u)

    grad = tree_map(partial(_jacrev_unravel, out), x, jac)

    return out, grad


def jax_vectorized_value_and_grad(f, x):
    out, pullback = vjp(f, x)

    ncols = sum([v.shape[1] for v in out.values()])

    u = dict()
    cols = dict()
    cnt = 0
    for k, v in out.items():
        if k not in cols:
            cols[k] = []

        n, m = v.shape
        a = np.zeros((ncols, n, m))
        for i in range(m):
            cols[k].append(cnt)
            a[cnt, :, i] = 1.0
            cnt += 1

        u[k] = anp.array(a)

    jac, = vmap(pullback, in_axes=0)(u)
    jac = np.array(jac)

    grad = {k: np.swapaxes(jac[I], 0, 1) for k, I in cols.items()}

    return out, grad



import numpy as np

from pymoo.problems.dyn import DynamicTestProblem
from pymoo.vendor.gta import fda2_deb, dMOP2, DIMP2, HE2, HE7, HE9


class FDA2_deb(DynamicTestProblem):

    def __init__(self, nt=10, taut=200):
        super().__init__(n_var=13,
                         n_obj=2,
                         nt=nt,
                         taut=taut)
        self.xl = -1.0 * np.ones(self.n_var)
        self.xl[0] = 0.0
        self.xu = 1.0 * np.ones(self.n_var)
        self.xu[0] = 1.0

        self.T_max = 200

    def _evaluate(self, X, out, *args, t=1, **kwargs):
        self.t = self.time(t)
        f1, f2 = fda2_deb(X, self.t)
        out["F"] = [f1, f2]

    def _calc_pareto_front(self, n_pareto_points=100):
        H = 0.75 + 0.75 * np.sin(0.5 * np.pi * self.t)
        H = 2 * np.sin(0.5 * np.pi * (self.t - 1))
        x = np.linspace(0.01, 1, n_pareto_points)
        # TODO Solve the inf result when x[0] = 0 and H(t)<0
        print(H, x[0], (1 - np.power(x, (1 / H)))[0], self.t)
        pf = np.column_stack([x, 1 - np.power(x, (H))])
        return pf


class DMOP2(DynamicTestProblem):

    def __init__(self, nt=10, taut=200):
        super().__init__(n_var=10,
                         n_obj=2,
                         nt=nt,
                         taut=taut)

    def _evaluate(self, X, out, *args, t=1, **kwargs):
        f1, f2 = dMOP2(X, self.time)
        out["F"] = [f1, f2]

    def _calc_pareto_front(self, n_pareto_points=100):
        H = 0.75 * np.sin(0.5 * np.pi * self.t) + 1.25
        x = np.linspace(0, 1, n_pareto_points)
        pf = np.column_stack([x, 1 - pow(x, H)])
        return pf


class DIMP2b(DynamicTestProblem):

    def __init__(self, nt=10, taut=200):
        super().__init__(n_var=10,
                         n_obj=2,
                         nt=nt,
                         taut=taut)
        self.xl = -2.0 * np.ones(self.n_var)
        self.xl[0] = 0.0
        self.xu = 2.0 * np.ones(self.n_var)
        self.xu[0] = 1.0

    def _evaluate(self, X, out, *args, t=1, **kwargs):
        self.t = self.time(t)
        f1, f2 = DIMP2(X, self.t)
        out["F"] = [f1, f2]

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0, 1, n_pareto_points)
        pf = np.column_stack([x, 1 - np.sqrt(x)])
        return pf

    def get_pf_t(self):
        return self._calc_pareto_front()


class HE2b(DynamicTestProblem):

    def __init__(self, nt=10, taut=200):
        super().__init__(n_var=30,
                         n_obj=2,
                         nt=nt,
                         taut=taut)

    def _evaluate(self, X, out, *args, t=1, **kwargs):
        self.t = self.time(t)
        f1, f2 = HE2(X, self.t)
        out["F"] = [f1, f2]

    def _calc_pareto_front(self, n_pareto_points=100):
        H = 0.75 * np.sin(0.5 * np.pi * self.t) + 1.25  # time may be incorrect
        x = np.linspace(0, 1, n_pareto_points)
        f2 = (1 - np.power(np.sqrt(x), H)) - (np.power(x, H) * np.sin(0.5 * np.pi * x))
        pf = np.column_stack([x, f2])
        return pf

    def get_pf_t(self):
        return self._calc_pareto_front()


class HE7b(DynamicTestProblem):

    def __init__(self, nt=10, taut=200):
        super().__init__(n_var=10,
                         n_obj=2,
                         nt=nt,
                         taut=taut)

        self.xl = -1 * np.ones(self.n_var)
        self.xl[0] = 0.0
        self.xu = 1 * np.ones(self.n_var)
        self.xu[0] = 1.0

    def time(self, t):
        self.tau = t
        t = 1 / self.nt
        t = t * np.floor(self.tau / self.taut)
        return t

    def _evaluate(self, X, out, *args, t=1, **kwargs):
        self.t = self.time(t)
        f1, f2 = HE7(X, self.t)
        out["F"] = [f1, f2]

    def _calc_pareto_front(self, n_pareto_points=100):
        H = 0.75 * np.sin(0.5 * np.pi * self.t) + 1.25  # current time may be incorrect
        x = np.linspace(0, 1, n_pareto_points)
        f2 = (2 - np.sqrt(x)) * (1 - np.power(x / (2 - np.sqrt(x)), H))
        pf = np.column_stack([x, f2])
        return pf

    def get_pf_t(self):
        return self._calc_pareto_front()


class HE9b(DynamicTestProblem):

    def __init__(self, nt=10, taut=200):
        super().__init__(n_var=10,
                         n_obj=2,
                         nt=nt,
                         taut=taut)

    def time(self, t):
        self.tau = t
        t = 1 / self.nt
        t = t * np.floor(self.tau / self.taut)
        return t

    def _evaluate(self, X, out, *args, t=1, **kwargs):
        self.t = self.time(t)
        f1, f2 = HE9(X, self.t)
        out["F"] = [f1, f2]

    def _calc_pareto_front(self, n_pareto_points=100):
        H = 0.75 * np.sin(0.5 * np.pi * self.t) + 1.25  # current time may be incorrect
        x = np.linspace(0, 1, n_pareto_points)
        f2 = (2 - np.sqrt(x)) * (1 - np.power(x / (2 - np.sqrt(x)), H))
        pf = np.column_stack([x, f2])
        return pf

    def get_pf_t(self):
        return self._calc_pareto_front()



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class Sphere(Problem):

    def __init__(self, n_var=10):
        super().__init__(n_var=n_var, n_obj=1, n_ieq_constr=0, xl=-0, xu=1, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = anp.sum(anp.square(x - 0.5), axis=1)

    def _calc_pareto_front(self):
        return 0.0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 0.5)
